# -*- coding: utf-8 -*-

from rst2pdf import createpdf
from docutils import writers
from StringIO import StringIO
import sys
from tempfile import mkstemp
import codecs

class PdfWriter(writers.Writer):

    def __init__(self, builder):
        writers.Writer.__init__(self)
        self.builder = builder
        self.output = u''

    supported = ('pdf')
    """Formats this writer supports."""

    config_section = 'pdf writer'
    config_section_dependencies = ('writers',)

    """Final translated form of `document`."""

    def translate(self):
        sio=StringIO('')
        createpdf.RstToPdf(sphinx=True).createPdf(doctree=self.document,output=sio,compressed=False)
        self.output=unicode(sio.getvalue(),'utf-8','ignore')

    def supports(self, format):
        """This writer supports all format-specific elements."""
        return 1



# -*- coding: utf-8 -*-
#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.8/rst2pdf/styles.py $
#$LastChangedDate: 2008-09-11 22:40:00 -0300 (Thu, 11 Sep 2008) $
#$LastChangedRevision: 338 $

import os
import sys
import re
from reportlab.platypus import *
import reportlab.lib.colors as colors
import reportlab.lib.units as units
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.lib.fonts import addMapping
from reportlab.lib.styles import *
from reportlab.lib.enums import *
from reportlab.pdfbase import pdfmetrics
import reportlab.lib.pagesizes as pagesizes
from log import log
from simplejson import loads
import docutils.nodes
import findfonts

try:
    from wordaxe.rl.paragraph import Paragraph
    from wordaxe.rl.styles import ParagraphStyle,getSampleStyleSheet
except:
    pass
    #log.warning("No hyphenation support, install wordaxe")

unit_separator = re.compile('(-?[0-9\.]*)')

class StyleSheet(object):
    '''Class to handle a collection of stylesheets'''
    def __init__(self,flist,fontPath=[]):
        log.info('Using stylesheets: %s'%','.join(flist))
        '''flist is a list of stylesheet filenames.They will
        be loaded and merged in order'''
        self.FontSearchPath=fontPath+['.',os.path.join(os.path.abspath(os.path.dirname(__file__)), 'fonts')]
        self.StyleSearchPath=['.',os.path.join(os.path.abspath(os.path.dirname(__file__)), 'styles')]

        findfonts.flist=self.FontSearchPath
        # Page width, height
        self.pw=0
        self.ph=0

        # Page size [w,h]
        self.ps=None

        # Margins (top,bottom,left,right,gutter)
        self.tm=0
        self.bm=0
        self.lm=0
        self.rm=0
        self.gm=0

        #text width
        self.tw=0

        self.languages=[]
        ssdata=[]
        # First parse all files and store the results
        for fname in flist:
            fname = self.findStyle(fname)
            try:
                ssdata.append(loads(open(fname).read()))
            except ValueError,e: # Error parsing the JSON data
                log.critical('Error parsing stylesheet "%s": %s'%(fname,str(e)))
                sys.exit(1)
            except IOError,e: #Error opening the ssheet
                log.critical('Error opening stylesheet "%s": %s'%(fname,str(e)))
                sys.exit(1)

        # Get pageSetup data from all stylessheets in order:
        self.ps=pagesizes.A4
        for data, ssname in zip(ssdata, flist):
            page=data.get('pageSetup', {})
            if page:
                if page.get('size', None): # A standard size
                    if page['size'] in pagesizes.__dict__:
                        self.ps=list(pagesizes.__dict__[page['size']])
                    else:
                        log.critical('Unknown page size %s in stylesheet %s' % (page['size'],ssname))
                        sys.exit(1)
                else: #A custom size
                    # The sizes are expressed in some unit.
                    # For example, 2cm is 2 centimeters, and we need
                    # to do 2*cm (cm comes from reportlab.lib.units)
                    if 'width' in page:
                        self.ps[0]=self.adjustUnits(page['width'])
                    if 'height' in page:
                        self.ps[1]=self.adjustUnits(page['height'])
                self.pw,self.ph=self.ps
                if 'margin-left' in page:
                    self.lm=self.adjustUnits(page['margin-left'])
                if 'margin-right' in page:
                    self.rm=self.adjustUnits(page['margin-right'])
                if 'margin-top' in page:
                    self.tm=self.adjustUnits(page['margin-top'])
                if 'margin-bottom' in page:
                    self.bm=self.adjustUnits(page['margin-bottom'])
                if 'margin-gutter' in page:
                    self.gm=self.adjustUnits(page['margin-gutter'])
                if 'firstTemplate' in page:
                    self.firstTemplate=page['firstTemplate']

                # tw is the text width.
                # We need it to calculate header-footer height
                # and compress literal blocks.
                self.tw=self.pw-self.lm-self.rm-self.gm

        # Get page templates from all stylesheets
        self.pageTemplates={}
        for data, ssname in zip(ssdata,flist):
            templates=data.get('pageTemplates',{})
            # templates is a dictionary of pageTemplates
            for key in templates:
                template=templates[key]
                # template is a dict.
                # template[´frames'] is a list of frames
                if key in self.pageTemplates:
                    self.pageTemplates[key].update(template)
                else:
                    self.pageTemplates[key]=template

        # Get font aliases from all stylesheets in order
        self.fontsAlias={}
        for data, ssname in zip(ssdata, flist):
            self.fontsAlias.update(data.get('fontsAlias',{}))

        self.embedded=[]
        # Embed all fonts indicated in all stylesheets
        for data, ssname in zip(ssdata, flist):
            embedded=data.get('embeddedFonts',[])

            for font in embedded:
                try:
                    if isinstance(font,unicode): # Just a font name, try to embed it
                        # See if we can find the font
                        fname,pos=findfonts.guessFont(font)
                        fontList=findfonts.autoEmbed(font)
                        if not fontList:
                            fontList=findfonts.autoEmbed(fname)
                        if fontList is not None:
                            self.embedded+=fontList
                            # Maybe the font we got is not called the same as the one we gave
                            # so check that out
                            suff=["","-Oblique","-Bold","-BoldOblique"]
                            if not fontList[0].startswith(font):
                                # We need to create font aliases, and use them
                                for fname,aliasname in zip(fontList,[font+suffix for suffix in suff]):
                                    self.fontsAlias[aliasname]=fname
                        continue
                    
                    # Each "font" is a list of four files, which will be used for
                    # regular / bold / italic / bold+italic versions of the font.
                    # If your font doesn't have one of them, just repeat the regular
                    # font.

                    # Example, using the Tuffy font from http://tulrich.com/fonts/
                    # "embeddedFonts" : [
                    #                    ["Tuffy.ttf","Tuffy_Bold.ttf","Tuffy_Italic.ttf","Tuffy_Bold_Italic.ttf"]
                    #                   ],

                    # The fonts will be registered with the file name, minus the extension.

                    if font[0].lower().endswith('.ttf'): # A True Type font
                        for variant in font:
                            pdfmetrics.registerFont(TTFont(str(variant.split('.')[0]), self.findFont(variant)))
                            self.embedded.append(str(variant.split('.')[0]))

                        # And map them all together
                        regular,bold,italic,bolditalic = [ variant.split('.')[0] for variant in font ]
                        addMapping(regular,0,0,regular)
                        addMapping(regular,0,1,italic)
                        addMapping(regular,1,0,bold)
                        addMapping(regular,1,1,bolditalic)
                    else: # A Type 1 font
                        # For type 1 fonts we require [FontName,regular,italic,bold,bolditalic]
                        # where each variant is a (pfbfile,afmfile) pair.
                        # For example, for the URW palladio from TeX:
                        # ["Palatino",("uplr8a.pfb","uplr8a.afm"),("uplri8a.pfb","uplri8a.afm"),("uplb8a.pfb","uplb8a.afm"),("uplbi8a.pfb","uplbi8a.afm")]
                        faceName=font[0]
                        regular=pdfmetrics.EmbeddedType1Face(*font[1])
                        italic=pdfmetrics.EmbeddedType1Face(*font[2])
                        bold=pdfmetrics.EmbeddedType1Face(*font[3])
                        bolditalic=pdfmetrics.EmbeddedType1Face(*font[4])
                        
                except Exception, e:
                    try:
                        fname=font[0].split('.')[0]
                        log.error("Error processing font %s: %s",fname,str(e))
                        log.error("Registering %s as Helvetica alias",fname)
                        self.fontsAlias[fname]='Helvetica'
                    except Exception,e:
                        log.critical("Error processing font %s: %s",fname,str(e))
                        sys.exit(1)

        # Go though all styles in all stylesheets and find all fontNames.
        # Then decide what to do with them
        for data, ssname in zip(ssdata, flist):
            styles = data.get('styles', {})
            for [skey, style] in styles:
                for key in style:
                    if key == 'fontName' or key.endswith('FontName'):
                        if style[key] in self.fontsAlias: # It´s an alias, replace it
                            style[key]=self.fontsAlias[style[key]]
                        if style[key] in self.embedded: # Embedded already, nothing to do
                            continue
                        if style[key] in ("Courier",
                                          "Courier-Bold",
                                          "Courier-BoldOblique",
                                          "Courier-Oblique",
                                          "Helvetica",
                                          "Helvetica-Bold",
                                          "Helvetica-BoldOblique",
                                          "Helvetica-Oblique",
                                          "Symbol",
                                          "Times-Bold",
                                          "Times-BoldItalic",
                                          "Times-Italic",
                                          "Times-Roman",
                                          "ZapfDingbats"): # Standard font, nothing to do
                            continue

                        # Now we need to do something
                        # See if we can find the font
                        fname,pos=findfonts.guessFont(style[key])
                        fontList=findfonts.autoEmbed(style[key])
                        if not fontList:
                            fontList=findfonts.autoEmbed(fname)                            
                        if fontList is not None:
                            self.embedded+=fontList
                            # Maybe the font we got is not called the same as the one we gave
                            # so check that out
                            suff=["","-Oblique","-Bold","-BoldOblique"]
                            if not fontList[0].startswith(style[key]):
                                # We need to create font aliases, and use them
                                for fname,aliasname in zip(fontList,[style[key]+suffix for suffix in suff]):
                                    self.fontsAlias[aliasname]=fname
                                style[key]=self.fontsAlias[style[key]+suff[pos]]
                        else:
                            log.error("Unknown font: \"%s\", replacing with Helvetica",style[key])
                            style[key]="Helvetica"
                            
        # Get styles from all stylesheets in order
        self.stylesheet = {}
        self.styles=[]
        self.linkColor = 'navy'
        for data, ssname in zip(ssdata, flist):
            self.linkColor = data.get('linkColor') or self.linkColor
            styles = data.get('styles', {})
            for [skey, style] in styles:
                sdict = {}
                # FIXME: this is done completely backwards
                for key in style:
                    # Handle color references by name
                    if key == 'color' or key.endswith('Color') and style[key]:
                        style[key]=formatColor(style[key])

                    # Handle alignment constants
                    elif key == 'alignment':
                        style[key]={'TA_LEFT':0, 'TA_CENTER':1, 'TA_CENTRE':1, 'TA_RIGHT':2, 'TA_JUSTIFY':4, 'DECIMAL':8}[style[key]]

                    elif key == 'language':
                        if not style[key] in self.languages:
                            self.languages.append(style[key])
                    # Make keys str instead of unicode (required by reportlab)
                    sdict[str(key)]=style[key]
                    sdict['name']=skey
                # If the style already exists, update it
                if skey in self.stylesheet:
                    self.stylesheet[skey].update(sdict)
                else: # New style
                    self.stylesheet[skey]=sdict
                    self.styles.append(sdict)
        # And create  reportlabs stylesheet
        self.StyleSheet = StyleSheet1()
        for s in self.styles:
            if 'parent' in s:
                if s['parent'] is None:
                    if s['name'] != 'base':
                        s['parent'] = self.StyleSheet['base']
                    else:
                        del(s['parent'])
                else:
                    s['parent'] = self.StyleSheet[s['parent']]
            else:
                if s['name'] != 'base':
                    s['parent'] = self.StyleSheet['base']

            # If the style has no bulletFontName but it has a fontName, set it
            if ('bulletFontName' not in s) and ('fontName' in s):
                s['bulletFontName']=s['fontName']

            # Adjust fontsize units
            if 'fontSize' not in s:
                s['fontSize']=s['parent'].fontSize
            elif 'parent' in s:
                # This means you can set the fontSize to "2cm" or to "150%" which
                # will be calculated relative to the parent style
                s['fontSize']=self.adjustUnits(s['fontSize'],s['parent'].fontSize)
            else:
                # If s has no parent, it's base, which has an explicit point size
                # by default and % makes no sense, but guess it as % of 10pt
                s['fontSize']=self.adjustUnits(s['fontSize'],10)

            # If the leading is not set, but the size is, set it
            if 'leading' not in s:
                s['leading']=1.2*s['fontSize']

            # If the bullet font size is not set, set it as fontSize
            if ('bulletFontSize' not in s) and ('fontSize' in s):
                s['bulletFontSize'] = s['fontSize']

            self.StyleSheet.add(ParagraphStyle(**s))
            
    def __getitem__(self,key):
        return self.StyleSheet[key]

    def findStyle(self,fn):
        '''Given a style filename, searches for it in StyleSearchPath
        and returns the real file name'''
        if not os.path.isabs(fn):
            for D in self.StyleSearchPath:
                tfn = os.path.join(D,fn)
                if os.path.isfile(tfn):
                    return str(tfn)
        return str(fn)

    def findFont(self,fn):
        '''Given a font filename, searches for it in FontSearchPath
        and returns the real file name'''
        if not os.path.isabs(fn):
            for D in self.FontSearchPath:
                tfn = os.path.join(D,fn)
                if os.path.isfile(tfn):
                    return str(tfn)
        return str(fn)

    def styleForNode(self,node):
        '''Returns the right default style for any kind of node.
        That usually means "bodytext", but for sidebars, for
        example, it's sidebar'''

        if isinstance(node,docutils.nodes.sidebar):
            return self['sidebar']
        return self['bodytext']

    def tstyleHead(self,rows=1):
        '''Returns a table style spec for a table header of `rows` based on the
        table-heading style from the stylesheet'''
        # This alignment thing is exactly backwards from the alignment for paragraphstyles
        alignment={0:'LEFT', 1:'CENTER', 1:'CENTRE', 2:'RIGHT', 4:'JUSTIFY',8:'DECIMAL'}[self['table-heading'].alignment]
        return [('BACKGROUND',(0,0),(-1,rows-1),self['table-heading'].backColor),
                ('ALIGN',(0,0),(-1,rows-1),alignment),
                ('TEXTCOLOR',(0,0),(-1,rows-1),self['table-heading'].textColor),
                ('FONT',(0,0),(-1,rows-1),self['table-heading'].fontName,
                                          self['table-heading'].fontSize,
                                          self['table-heading'].leading),
                ('VALIGN',(0,0),(-1,rows-1),self['table-heading'].valign)
               ]
               
    def tstyleBody(self,rows=1):
        '''Returns a table style spec for a table of any size based on the
        table style from the stylesheet'''
        return [("ROWBACKGROUNDS",(0,0),(-1,-1),
                [formatColor(c,numeric=False) for c in self['table'].rowBackgrounds]),
            ]

    def pStyleToTStyle(self,style,x,y):
        '''Given a reportlab paragraph style, returns a spec for a table style
           that adopts some of its features (example, background color)'''
        results=[]
        if style.backColor:
            results.append(('BACKGROUND',(x,y),(x,y),style.backColor))
        if style.borderWidth:
            bw=style.borderWidth
            del(style.__dict__['borderWidth'])
            if style.borderColor:
                bc=style.borderColor
                del(style.__dict__['borderColor'])
            else:
                bc=colors.black
            results.append(('BOX',(x,y),(x,y),bw,bc))
        return results

    def adjustUnits(self, v, total = None):
        if total is None:
            total = self.pw
        return adjustUnits(v,total)
        
def adjustUnits(v,total=None):
    '''Takes something like 2cm and returns 2*cm.
    If you use % as a unit, it returns the percentage
    of "total".

    If total is not given, returns a percentage of the page
    width. However, if you get to that stage, you are
    doing it wrong.

    Example::

            >>> adjustUnits('50%',200)
            100
    '''

    v = str(v)
    _,n,u = re.split('(-?[0-9\.]*)',v)
    if not u:
        return float(n) # assume points
    if u in units.__dict__:
        return float(n) * units.__dict__[u]
    else:
        if u=='%':
            return float(n) * total/100
        log.error('Unknown unit "%s"' % u)
    return float(n)

               
# Some table styles used for pieces of the document

tstyles={}

# Used for regular tables

tstyleNorm = [ ('VALIGN',(0,0),(-1,-1),'TOP'),
               ('INNERGRID', (0,0), (-1,-1), 0.25, colors.black),
               ('BOX', (0,0), (-1,-1), 0.25, colors.black),
             ]

# Header row in tables


tstyles['normal']=TableStyle(tstyleNorm)

# Used for field lists

tstyles['field']=TableStyle([ ('VALIGN',(0,0),(-1,-1),'TOP'),
                              ('ALIGNMENT',(0,0),(1,-1),'RIGHT'),
                            ])
fieldlist_lwidth=3*units.cm

# Used for endnotes

tstyles['endnote']=TableStyle([ ('VALIGN',(0,0),(-1,-1),'TOP'),
                                ('ALIGNMENT',(0,0),(1,-1),'RIGHT'),
                              ])
# Left column of the endnote. The content of the note takes the rest of
# the available space
endnote_lwidth=2*units.cm

# Used for sidebars

tstyles['sidebar']=TableStyle([ ('VALIGN',(0,0),(-1,-1),'TOP'),
                             ('BACKGROUND',(0,0),(-1,-1),colors.lightyellow),
                           ])


tstyles['bullet']=TableStyle([ ('VALIGN',(0,0),(-1,-1),'TOP'),
                             ])

def formatColor(value,numeric=True):
    '''Converts a color like "gray" or "0xf" or "ffff" to something
    reportlab will like'''
    if value in colors.__dict__:
        return colors.__dict__[value]
    else: # Hopefully, a hex color:
        c = value.strip()
        if c[0] == '#':
            c = c[1:]
        while len(c) < 6:
            c = '0'+c
        if numeric:
            r = int(c[:2],16)
            g = int(c[2:4],16)
            b = int(c[4:6],16)
            return colors.Color(r,g,b)
        else:
            return str("#"+c)



# -*- coding: utf-8 -*-
#$HeadURL: https://rst2pdf.googlecode.com/svn/trunk/rst2pdf/tests/test_rst2pdf.py $
#$LastChangedDate: 2008-08-29 16:09:08 +0200 (Fri, 29 Aug 2008) $
#$LastChangedRevision: 160 $
from unittest import TestCase, makeSuite

from docutils.core import publish_doctree
import rst2pdf
#import pdb; pdb.set_trace()

from os.path import join
from utils import PREFIX

import pdb


def input_file_path(file):
    """ unused here
    is looked for in tests below
    where am I"""
    return join(PREFIX, 'input', file)

class IncludeTests(TestCase):
    def test_wrong_file(self):
        input="""
This one gives a warning, non existent file:

.. code-block:: python
   :include: xyzzy.py
"""
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'literal_block')
        self.assertEqual(include.astext(), u'\n')
        
    def test_missing_file(self):
        input="""
This one gives a warning, missing file:

.. code-block:: python
   :include:
"""
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'system_message')
        self.assertEqual(include.children[1].tagname, 'literal_block')
        self.assertEqual(include.children[1].astext(), u'.. code-block:: python\n   :include:')

    def test_existing_file(self):
        input="""
This one exists:

.. code-block:: py
   :include: %s
""" % __file__
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'literal_block')
        self.assertEqual(include.children[0].astext(), u'# -*- coding: utf-8 -*-')

    def test_wrong_lang(self):
        input="""
This one exists:

.. code-block:: nothing
   :include: %s
""" % __file__
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'literal_block')
        self.assertEqual(include.astext().split('\n')[0], u'# -*- coding: utf-8 -*-')

    def test_existing_file_start_at(self):
        input="""
This one exists:

.. code-block:: py
   :include: %s
   :start-at: def input_file_path(file):
""" % __file__
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'literal_block')
        self.assertEqual(include.astext().split('\n')[0], u'def input_file_path(file):')
        self.assertEqual(include.astext().split('\n')[-2:][0], u"    unittest.main(defaultTest='test_suite')")
        
    def test_existing_file_start_after(self):
        input="""
This one exists:

.. code-block:: py
   :include: %s
   :start-after: def input_file_path(file):
""" % __file__
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'literal_block')
        self.assertEqual(include.astext().split('\n')[0], u'    """ unused here')
        self.assertEqual(include.astext().split('\n')[-2:][0], u"    unittest.main(defaultTest='test_suite')")
        
    def test_existing_file_end_before(self):
        input="""
This one exists:

.. code-block:: py
   :include: %s
   :end-before: return join(PREFIX, 'input', file)
""" % __file__
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'literal_block')
        self.assertEqual(include.children[0].astext(), u'# -*- coding: utf-8 -*-')
        self.assertEqual(include.astext().split('\n')[-3:][0], u'    where am I"""')
        
    def test_existing_file_end_at(self):
        input="""
This one exists:

.. code-block:: py
   :include: %s
   :end-at: def input_file_path(file):
""" % __file__
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'literal_block')
        self.assertEqual(include.children[0].astext(), u'# -*- coding: utf-8 -*-')
        self.assertEqual(include.astext().split('\n')[-2:][0], u'def input_file_path(file):')
        
    def test_existing_file_start_at_end_at(self):
        input="""
This one exists:

.. code-block:: py
   :include: %s
   :start-at: def input_file_path(file):
   :end-at: return join(PREFIX, 'input', file)
""" % __file__
        doctree=publish_doctree(input)
        include = doctree.children[1]
        self.assertEqual(include.tagname, 'literal_block')
        self.assertEqual(include.astext().split('\n')[0], u'def input_file_path(file):')
        self.assertEqual(include.astext().split('\n')[-2:][0], u"    return join(PREFIX, 'input', file)")
        
        
def test_suite():
    suite = makeSuite(IncludeTests)
    return suite
    
if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')


# -*- coding: utf-8 -*-
#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.8/rst2pdf/tests/test_rst2pdf.py $
#$LastChangedDate: 2008-09-03 14:48:33 -0300 (Wed, 03 Sep 2008) $
#$LastChangedRevision: 264 $
from unittest import TestCase, makeSuite
from pyPdf import PdfFileReader
import cStringIO

from docutils.core import publish_doctree
import reportlab

import rst2pdf
from utils import *

import pdb
#import pdb; pdb.set_trace()

def input_file_path(file):
    return join(PREFIX, 'input', file)

class FullGenerationTests(TestCase):
    """ broken since createPdf is no longer a function
    """
    def test_bullet_chars_full(self):
        self.output=cStringIO.StringIO()
        input_file = input_file_path('test_bullet_chars.txt')
        input=open(input_file,'r').read()
        createPdf(text=input, output=self.output, styleSheet=None)
        self.output.seek(0)
        result = PdfFileReader(self.output)
        text_result = result.getPage(0).extractText()
        self.assertEqual(text_result, u'\nTest\nItem 1\nItem 2\n')
        
    
class GenerationTests(rst2pdfTests):

        
    def test_transition(self):
        input="""
Transitions
-----------

Here's a transition:

---------

It divides the section.
"""
        doctree=publish_doctree(input)
        elements=self.converter.gen_elements(doctree,0)
        self.assertEqual(len(elements), 5)
        assert(isinstance(elements[3],rst2pdf.flowables.Separation))
        
    def test_strong(self):
        input="""
**strong**"""
        doctree=publish_doctree(input)
        elements=self.converter.gen_elements(doctree,0)
        self.assertEqual(len(elements), 1)
        para = elements[0]
        self.assertEqual(para.text, '<b>strong</b>')
        parafrag = para.frags[0]
        self.assertEqual(parafrag.bold, 1)
        
    def test_emphasis(self):
        input="""
*emphasis*"""
        doctree=publish_doctree(input)
        elements=self.converter.gen_elements(doctree,0)
        self.assertEqual(len(elements), 1)
        para = elements[0]
        self.assertEqual(para.text, '<i>emphasis</i>')
        parafrag = para.frags[0]
        self.assertEqual(parafrag.italic, 1)
        # pdb.set_trace()
        
    def test_raw_pagebreak(self):
        input="""
One page

.. raw:: pdf

   PageBreak

Another page.
"""
        doctree=publish_doctree(input)
        elements=self.converter.gen_elements(doctree,0)
        self.assertEqual(len(elements), 3)
        pagebreak = elements[1]
        self.assertEqual(pagebreak.__class__, rst2pdf.flowables.MyPageBreak)
        #pdb.set_trace()
        
    def test_raw_spacer(self):
        input="""
A paragraph

.. raw:: pdf

    Spacer 0,200

And another paragraph.

"""
        doctree=publish_doctree(input)
        elements=self.converter.gen_elements(doctree,0)
        self.assertEqual(len(elements), 3)
        spacer = elements[1]
        self.assertEqual(spacer.__class__, reportlab.platypus.flowables.Spacer)
        #pdb.set_trace()
        
    def test_bullets(self):
        input="""
 Test
======

- Item 1
- Item 2
"""
        doctree=publish_doctree(input)
        elements=self.converter.gen_elements(doctree,0)
        self.assertEqual(len(elements), 4)
        self.assertEqual(elements[1].text, 'Test')
        self.assertEqual(elements[2]._cellvalues[0][0][0].text, 'Item 1')
        self.assertEqual(elements[2]._cellvalues[0][0][0].bulletText, u'\u2022')
        self.assertEqual(elements[3]._cellvalues[0][0][0].text, 'Item 2')
        self.assertEqual(elements[3]._cellvalues[0][0][0].bulletText, u'\u2022')
        
    def test_sidebar(self):
        input="""
.. sidebar:: The Sidebar

   This is a real sidebar, declared with the sidebar directive.

Quisque dignissim. Duis in velit vel augue rhoncus pretium. Duis non nisl in lorem placerat rutrum. 
"""
        doctree=publish_doctree(input)
        elements=self.converter.gen_elements(doctree,0)
        self.assertEqual(len(elements), 2)
        sidebar = elements[0]
        self.assertEqual(sidebar.__class__, rst2pdf.flowables.Sidebar)
        #pdb.set_trace()

        
def test_suite():
    suite = makeSuite(GenerationTests)
    return suite
    
if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')


# -*- coding: utf-8 -*-
#$HeadURL: https://rst2pdf.googlecode.com/svn/trunk/rst2pdf/tests/test_rst2pdf.py $
#$LastChangedDate: 2008-08-29 16:09:08 +0200 (Fri, 29 Aug 2008) $
#$LastChangedRevision: 160 $

from unittest import TestCase
from os.path import join, abspath, dirname, basename
PREFIX = abspath(dirname(__file__))

from rst2pdf.createpdf import RstToPdf

def input_file_path(file):
    return join(PREFIX, 'input', file)

class rst2pdfTests(TestCase):

    def setUp(self):
        self.converter=RstToPdf()



#!/usr/bin/env python

'''Scan a list of folders and find all .afm files, then create rst2pdf-ready font-aliases.'''

import os,sys
from log import log
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.lib.fonts import addMapping

flist=[]
afmList=[]
pfbList={}

# Aliases defined by GhostScript, so if you use Palatino or whatever you
# may get **something**. They are family name aliases.
Alias = {
    'ITC Bookman'            : 'URW Bookman L',
    'ITC Avant Garde Gothic' : 'URW Gothic L',
    'Palatino'               : 'URW Palladio L',
    'New Century Schoolbook' : 'Century Schoolbook L',
    'ITC Zapf Chancery'      : 'URW Chancery L'
    }

# Standard PDF fonts, so no need to embed them
Ignored = ['Times', 'ITC Zapf Dingbats', 'Symbol', 'Helvetica', 'Courier']



fonts={}
families={}
fontMappings={}


def loadFonts():
    '''Given a font name, returns the actual font files. If the font name is not valid, it will treat it as a
    font family name, and return the value of searching for the regular font of said family.'''    
    if not afmList or not pfbList:
        # Find all ".afm" and ".pfb" files files
        def findFontFiles(_,folder,names):
            for f in os.listdir(folder):
                if f.lower().endswith('.afm'):
                    afmList.append(os.path.join(folder,f))
                if f.lower().endswith('.pfb'):
                    pfbList[f[:-4]]=os.path.join(folder,f)

        for folder in flist:
            os.path.walk(folder,findFontFiles,None)

        # Now we have full afm and pbf lists, process the afm list to figure out
        # family name weight and if it's italic or not, as well as where the
        # matching pfb file is

        for afm in afmList:
            family=None
            fontName=None
            italic=False
            bold=False
            pfb=None
            for line in open(afm,'r'):
                line=line.strip()
                if line.startswith('StartCharMetrics'):
                    break
                elif line.startswith('FamilyName'):
                    family=' '.join(line.split(' ')[1:])
                elif line.startswith('FontName'):
                    fontName=line.split(' ')[1]
                elif line.startswith('Weight'):
                    w=line.split(' ')[1]
                    if w=='Bold':
                        bold=True
                elif line.startswith('ItalicAngle'):
                    if line.split(' ')[1]<>'0.0':
                        italic=True

            baseName=os.path.basename(afm)[:-4]
            if family in Ignored:
                continue
            if family in Alias:
                continue
            if baseName not in pfbList:
                #print "Warning: afm file without matching pfb file: %s"%baseName
                continue

            # So now we have a font we knopw we can embed.
            fonts[fontName]=(afm,pfbList[baseName],family)

            # And we can try to build/fill the family mapping
            if family not in families:
                families[family]=[fontName,fontName,fontName,fontName]
            if bold and italic: families[family][3]=fontName
            elif bold: families[family][2]=fontName
            elif italic: families[family][1]=fontName
            # FIXME: what happens if there are Demi and Medium weights? We get a random one.
            else: families[family][0]=fontName

def findFont(fname):
    loadFonts()
    # So now we are sure we know the families and font names. Well, return some data!
    if fname in fonts:
        font=fonts[fname]
    else:
        if fname in Alias:
            fname=Alias[fname]
        if fname in families:
            font=fonts[families[fname][0]]
        else:
            return None
    return font

def findTTFont(fname):
    def get_family(query):
        data=os.popen("fc-match \"%s\""%query,"r").read()
        for line in data.splitlines():
            line=line.strip()
            if not line: continue
            fname,family,_,variant=line.split('"')[:4]
            family=family.replace('"','')
            if family:
                return family
        return None

    def get_fname(query):
        data=os.popen("fc-match -v \"%s\""%query,"r").read()
        for line in data.splitlines():
            line=line.strip()
            if line.startswith("file: "):
                return line.split('"')[1]
        return None

    def get_variants(family):
        variants=[ get_fname(family+":style=Roman"),
                 get_fname(family+":style=Oblique"),
                 get_fname(family+":style=Bold"),
                 get_fname(family+":style=Bold Oblique")]
        if variants[1]==variants[0]:
            variants[1]=get_fname(family+":style=Italic")
        if variants[3]==variants[0]:
            variants[3]=get_fname(family+":style=Bold Italic")
        if variants[0].endswith('.pfb') or variants[0].endswith('.gz') :
            return None
        return variants
        
    family=get_family(fname)
    if not family:
        log.error("Unknown font: %s",fname)
        return None
    return get_variants(family)

def autoEmbed(fname):
    '''Given a font name, does a best-effort of embedding said font
    and its variants, returning a list of the font names
    it registered with reportlab'''
    fontList=[]
    f=findFont(fname)
    if f: # It's a Type 1 font, and we have it
        family=families[f[2]]

        # Register the whole family of faces
        faces= [ pdfmetrics.EmbeddedType1Face(*fonts[fn][:2]) \
                    for fn in family]
        for face in faces:
            pdfmetrics.registerTypeFace(face)

        for face,name in zip(faces,family):
            fontList.append(name)
            font=pdfmetrics.Font(face,name,"WinAnsiEncoding")
            pdfmetrics.registerFont(font)

        # Map the variants
        regular,italic,bold,bolditalic=family
        addMapping(fname,0,0,regular)
        addMapping(fname,0,1,italic)
        addMapping(fname,1,0,bold)
        addMapping(fname,1,1,bolditalic)
        addMapping(regular,0,0,regular)
        addMapping(regular,0,1,italic)
        addMapping(regular,1,0,bold)
        addMapping(regular,1,1,bolditalic)
        return fontList

    variants=findTTFont(fname)
    if variants: #It is a TT Font and we found it using fc-match (or found *something*)
        for variant in variants:
            vname=os.path.basename(variant)[:-4]
            pdfmetrics.registerFont(TTFont(vname,variant))
            fontList.append(vname)
        ## Also register "standard" aliases for the variants
        #for variant,name in zip(variants,[fname+suffix for suffix in ["","-Oblique","-Bold","-BoldOblique"]]):
            #fontList.append(name)
            #pdfmetrics.registerFont(TTFont(name,variant))

        # And map them all together
        regular,bold,italic,bolditalic = [ os.path.basename(variant)[:-4] for variant in variants ]
        addMapping(regular,0,0,regular)
        addMapping(regular,0,1,italic)
        addMapping(regular,1,0,bold)
        addMapping(regular,1,1,bolditalic)
    return fontList


def guessFont(fname):
    '''given a font name (like Tahoma-BoldOblique) guess what it means and return
    (family,x)

    where x is 0: regular
               1: italic
               2: bold
               3: bolditalic

    May beuseful someday.
    '''

    if '-' not in fname:
        return fname,0
    italic=0
    bold=0
    family,mod=fname.split('-')
    mod=mod.lower()
    if "oblique" in mod or "italic" in mod:
        italic=1
    if "bold" in mod or "black" in mod:
        bold=1
    return family,2*bold+italic
            

def main():
    global flist
    if len(sys.argv)<>2:
        print "Usage: findfont fontName"
        sys.exit(1)
    flist=["/usr/share/fonts","/usr/share/texmf-dist/fonts"]
    f=findFont(sys.argv[1])
    if not f:
        f=findTTFont(sys.argv[1])
    if f:
        print f
    else:
        fn,pos=guessFont(sys.argv[1])
        f=findFont(fn)
        if not f:
            f=findTTFont(fn)
        if f:
            print f
        else:
            print "Unknown font %s"%sys.argv[1]

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import logging,sys

log = logging.getLogger('rst2pdf')
hdlr = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('[%(levelname)s] %(filename)s L%(lineno)d %(message)s')
hdlr.setFormatter(formatter)
log.addHandler(hdlr)
log.setLevel(logging.WARNING)



# -*- coding: utf-8 -*-

# Copyright (C) 2007-2008 by Igor Novikov
# Copyright (C) 2000, 2001, 2002 by Bernhard Herzog
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

###Sketch Config
#type = Export
#tk_file_type = ("Portable Document Format (PDF)", '.pdf')
#extensions = '.pdf'
#format_name = 'PDF-Reportlab'
#unload = 1
###End

from math import atan2, pi
import PIL
from app import _,Bezier, EmptyPattern, Rotation, Translation, _sketch
from app.Graphics.curveop import arrow_trafos
import reportlab.pdfgen.canvas
import app


def make_pdf_path(pdfpath, paths):
	for path in paths:
		for i in range(path.len):
			type, control, p, cont = path.Segment(i)
			if type == Bezier:
				p1, p2 = control
				pdfpath.curveTo(p1.x, p1.y, p2.x, p2.y, p.x, p.y)
			else:
				if i > 0:
					pdfpath.lineTo(p.x, p.y)
				else:
					pdfpath.moveTo(p.x, p.y)
		if path.closed:
			pdfpath.close()
	return pdfpath

class PDFDevice:

	has_axial_gradient = 0
	has_radial_gradient = 0
	has_conical_gradient = 0
	gradient_steps = 100

	def __init__(self, pdf):
		self.pdf = pdf

	def PushTrafo(self):
		self.pdf.saveState()

	def Concat(self, trafo):
		apply(self.pdf.transform, trafo.coeff())

	def Translate(self, x, y = None):
		if y is None:
			x, y = x
		self.pdf.translate(x, y)

	def Rotate(self, angle):
		self.pdf.rotate(angle)

	def Scale(self, scale):
		self.pdf.scale(scale, scale)
	
	def PopTrafo(self):
		self.pdf.restoreState()

	PushClip = PushTrafo
	PopClip = PopTrafo

	def SetFillColor(self, color):
		self.pdf.setFillColor(tuple(color))

	def SetLineColor(self, color):
		self.pdf.setStrokeColor(tuple(color))

	def SetLineAttributes(self, width, cap = 1, join = 0, dashes = ()):
		self.pdf.setLineWidth(width)
		self.pdf.setLineCap(cap - 1)
		self.pdf.setLineJoin(join)
		if dashes:
			dashes = list(dashes)
			w = width
			if w < 1.0:
				w = 1.0
			for i in range(len(dashes)):
				dashes[i] = w * dashes[i]
		self.pdf.setDash(dashes)

	def DrawLine(self, start, end):
		self.pdf.line(start.x, start.y, end.x, end.y)

	def DrawLineXY(self, x1, y1, x2, y2):
		self.pdf.line(x1, y1, x2, y2)

	def DrawRectangle(self, start, end):
		self.pdf.rectangle(start.x, start.y, end.x - start.x, end.y - start.y, 
						   1, 0)

	def FillRectangle(self, left, bottom, right, top):
		self.pdf.rect(left, bottom, right - left, top - bottom, 0, 1)

	def DrawCircle(self, center, radius):
		self.pdf.circle(center.x, center.y, radius, 1, 0)

	def FillCircle(self, center, radius):
		self.pdf.circle(center.x, center.y, radius, 0, 1)

	def FillPolygon(self, pts):
		path = self.pdf.beginPath()
		apply(path.moveTo, pts[0])
		for x, y in pts:
			path.lineTo(x, y)
		path.close()
		self.pdf.drawPath(path, 0, 1)

	def DrawBezierPath(self, path, rect = None):
		self.pdf.drawPath(make_pdf_path(self.pdf.beginPath(), (path,)), 1, 0)

	def FillBezierPath(self, path, rect = None):
		self.pdf.drawPath(make_pdf_path(self.pdf.beginPath(), (path,)), 0, 1)



class PDFGenSaver:

	def __init__(self, file, filename, document, options):
		self.file = file
		self.filename = filename
		self.document = document
		self.options = options

		# if there's a pdfgen_canvas option assume it's an instance of
		# reportlab.pdfgen.canvas.Canvas that we should render on. This
		# allows multiple documents to be rendered into the same PDF
		# file or to have other python code outside of Sketch such as
		# reportlab itself (more precisely one of its other components
		# besides pdfgen) render into to too.
		#
		# The code here assumes that the canvas is already setup
		# properly.
		if options.has_key("pdfgen_canvas"):
			self.pdf = options["pdfgen_canvas"]
		else:
			self.pdf = reportlab.pdfgen.canvas.Canvas(file)
			self.pdf.setPageSize(document.PageSize())

	def close(self):
		if not self.options.has_key("pdfgen_canvas"):
			self.pdf.save()

	def set_properties(self, properties, bounding_rect = None):
		pattern = properties.line_pattern
		if not pattern.is_Empty:
			if pattern.is_Solid:
				c, m, y, k =pattern.Color().getCMYK()
				self.pdf.setStrokeColorCMYK(c, m, y, k)
			self.pdf.setLineWidth(properties.line_width)
			self.pdf.setLineJoin(properties.line_join)
			self.pdf.setLineCap(properties.line_cap - 1)
			dashes = properties.line_dashes
			if dashes:
				dashes = list(dashes)
				w = properties.line_width
				if w < 1.0:
					w = 1.0
				for i in range(len(dashes)):
					dashes[i] = w * dashes[i]
			self.pdf.setDash(dashes)
		active_fill = None
		pattern = properties.fill_pattern
		if not pattern.is_Empty:
			if pattern.is_Solid:
				c, m, y, k =pattern.Color().getCMYK()
				self.pdf.setFillColorCMYK(c, m, y, k)
			elif pattern.is_Tiled:
				pass
			elif pattern.is_AxialGradient:
				active_fill = self.axial_gradient
			else:
				active_fill = self.execute_pattern
		return active_fill

	def axial_gradient(self, properties, rect):
		pattern = properties.fill_pattern
		vx, vy = pattern.Direction()
		angle = atan2(vy, vx) - pi / 2
		center = rect.center()
		rot = Rotation(angle, center)
		left, bottom, right, top = rot(rect)
		trafo = rot(Translation(center))
		image = PIL.Image.new('RGB', (1, 200))
		border = int(round(100 * pattern.Border()))
		_sketch.fill_axial_gradient(image.im, pattern.Gradient().Colors(), 
									0, border, 0, 200 - border)
		self.pdf.saveState()
		apply(self.pdf.transform, trafo.coeff())
		self.pdf.drawInlineImage(image, (left - right) / 2, (bottom - top) / 2, 
								 right - left, top - bottom)
		self.pdf.restoreState()

	def execute_pattern(self, properties, rect):
		device = PDFDevice(self.pdf)
		properties.fill_pattern.Execute(device, rect)

	def make_pdf_path(self, paths):
		return make_pdf_path(self.pdf.beginPath(), paths)

	def polybezier(self, paths, properties, bounding_rect, clip = 0):
		pdfpath = self.make_pdf_path(paths)
		active_fill = self.set_properties(properties, bounding_rect)
		if active_fill:
			if not clip:
				self.pdf.saveState()
			self.pdf.clipPath(pdfpath, 0, 0)
			active_fill(properties, bounding_rect)
			if not clip:
				self.pdf.restoreState()
			if properties.HasLine():
				self.pdf.drawPath(pdfpath, 1, 0)
		else:
			if clip:
				method = self.pdf.clipPath
			else:
				method = self.pdf.drawPath
			method(self.make_pdf_path(paths), properties.HasLine(), 
				   properties.HasFill())
		# draw the arrows
		if properties.HasLine():
			# Set the pdf fill color to the line color to make sure that
			# arrows that are filled are filled with the line color of
			# the object. Since lines are always drawn last, this
			# shouldn't interfere with the object's fill.
			c, m, y, k = properties.line_pattern.Color().getCMYK()
			self.pdf.setFillColorCMYK(c, m, y, k)
			arrow1 = properties.line_arrow1
			arrow2 = properties.line_arrow2
			if arrow1 or arrow2:
				for path in paths:
					t1, t2 = arrow_trafos(path, properties)
					if arrow1 and t1 is not None:
						self.draw_arrow(arrow1, t1)
					if arrow2 and t2 is not None:
						self.draw_arrow(arrow2, t2)
					
	def draw_arrow(self, arrow, trafo):
		path = arrow.Paths()[0].Duplicate()
		path.Transform(trafo)
		pdfpath = self.make_pdf_path((path,))
		if arrow.IsFilled():
			self.pdf.drawPath(pdfpath, 0, 1)
		else:
			self.pdf.drawPath(pdfpath, 1, 0)

	def mask_group(self, object):
		mask = object.Mask()
		if not mask.has_properties:
			# XXX implement this case (raster images)
			return
		if mask.is_curve:
			self.pdf.saveState()
			prop = mask.Properties().Duplicate()
			prop.SetProperty(line_pattern = EmptyPattern)
			self.polybezier(mask.Paths(), prop, mask.bounding_rect, clip = 1)
			self.save_objects(object.MaskedObjects())
			if mask.has_line and mask.Properties().HasLine():
				prop = mask.Properties().Duplicate()
				prop.SetProperty(fill_pattern = EmptyPattern)
				self.polybezier(mask.Paths(), prop, mask.bounding_rect, 
								clip = 1)
			self.pdf.restoreState()

	def raster_image(self, object):
		self.pdf.saveState()
		apply(self.pdf.transform, object.Trafo().coeff())
		self.pdf.drawInlineImage(object.Data().Image(), 0, 0)
		self.pdf.restoreState()

	def simple_text(self, object, clip = 0):
		properties = object.Properties()
		active_fill = self.set_properties(properties, object.bounding_rect)
		fontname = properties.font.PostScriptName()
		if fontname not in self.pdf.getAvailableFonts():
			fontname = 'Times-Roman'

		if active_fill and not clip:
			self.pdf.saveState()
		pdftext = self.pdf.beginText()
		if active_fill:
			pdftext.setTextRenderMode(7)
		elif clip:
			pdftext.setTextRenderMode(4)
		pdftext.setFont(fontname, properties.font_size)
		apply(pdftext.setTextTransform, object.FullTrafo().coeff())
		pdftext.textOut(object.Text())
		self.pdf.drawText(pdftext)
		if active_fill:
			active_fill(properties, object.bounding_rect)
			if not clip:
				self.pdf.restoreState()

	def path_text(self, object, clip = 0):
		properties = object.Properties()
		active_fill = self.set_properties(properties, object.bounding_rect)
		fontname = properties.font.PostScriptName()
		if fontname not in self.pdf.getAvailableFonts():
			fontname = 'Times-Roman'

		if active_fill and not clip:
			self.pdf.saveState()
		pdftext = self.pdf.beginText()
		if active_fill:
			pdftext.setTextRenderMode(7)
		elif clip:
			pdftext.setTextRenderMode(4)
		pdftext.setFont(fontname, properties.font_size)
		trafos = object.CharacterTransformations()
		text = object.Text()
		for i in range(len(trafos)):
			apply(pdftext.setTextTransform, trafos[i].coeff())
			pdftext.textOut(text[i])
		self.pdf.drawText(pdftext)
		if active_fill:
			active_fill(properties, object.bounding_rect)
			if not clip:
				self.pdf.restoreState()

	def Save(self):
		self.document.updateActivePage()
		masters=self.document.getMasterLayers()
		count=0
		pagenum=len(self.document.pages)
		interval=int(97/pagenum)
		for page in self.document.pages:
			count+=1
			app.updateInfo(inf2=_('Composing page %u of %u')%(count,pagenum),inf3=count*interval)
			layers=page+masters
			for layer in layers:
				if not layer.is_SpecialLayer and layer.Printable():
					self.save_objects(layer.GetObjects())
			#self.pdf.showPage()

	def save_objects(self, objects):
		for object in objects:
			if object.is_Compound:
				if object.is_MaskGroup:
					self.mask_group(object)
				else:
					self.save_objects(object.GetObjects())
			elif object.is_SimpleText:
#				self.simple_text(object)
				obj=object.AsBezier()
				self.polybezier(obj.Paths(), obj.Properties(), obj.bounding_rect)
			elif object.is_PathTextText:
				self.path_text(object)				
			elif object.is_Image:
				self.raster_image(object)
			elif object.is_Bezier or object.is_Rectangle or object.is_Ellipse:
				self.polybezier(object.Paths(), object.Properties(), object.bounding_rect)



def save(document, file, filename, options = {}):
	app.updateInfo(inf1=_('PDF generation.'),inf2=_('Start document composing'),inf3=3)
	saver = PDFGenSaver(file, filename, document, options)
	saver.Save()
	saver.close()
	app.updateInfo(inf2=_('Document generation is finished'),inf3=100)



# -*- coding: utf-8 -*-

import sys,os
from reportlab.platypus import *
from log import log

HAS_UNICONVERTOR=False

try:
    for p in sys.path:
        d=os.path.join(p,'uniconvertor')
        if os.path.isdir(d):
            sys.path.append(d)
            from app.io import load
            from app.plugins import plugins
            import app
            from uniconvsaver import save
            app.init_lib()
            plugins.load_plugin_configuration()
            HAS_UNICONVERTOR=True
            break
except ImportError:
    pass

class SVGImage(Flowable):
    def __init__(self,filename,width=None,height=None):
        Flowable.__init__(self)
        if HAS_UNICONVERTOR:
            self.doc = load.load_drawing(filename)
            self.saver = plugins.find_export_plugin(plugins.guess_export_plugin(".pdf"))
            self.width=width
            self.height=height
            _,_,self._w,self._h=self.doc.BoundingRect()
            if not self.width: self.width=self._w
            if not self.height: self.height=self._h
        else:
            log.error("SVG image support not enabled, install uniconvertor")
    
    def wrap(self,aW,aH):
        if HAS_UNICONVERTOR:
            return self.width ,self.height
        return 0,0
    
    def drawOn(self,canv,x,y,_sW=0):
        if HAS_UNICONVERTOR:

            if _sW and hasattr(self,'hAlign'):
                from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT, TA_JUSTIFY
                a = self.hAlign
                if a in ('CENTER','CENTRE', TA_CENTER):
                    x = x + 0.5*_sW
                elif a in ('RIGHT',TA_RIGHT):
                    x = x + _sW
                elif a not in ('LEFT',TA_LEFT):
                    raise ValueError, "Bad hAlign value "+str(a)
        
            canv.saveState()
            canv.translate(x,y)
            canv.scale(self.width/self._w,self.height/self._h)
            save (self.doc,open(".ignoreme.pdf","w"),".ignoreme.pdf",options={'pdfgen_canvas':canv})
            os.unlink(".ignoreme.pdf")
            canv.restoreState()
           
           
if __name__ == "__main__":
    from reportlab.lib.styles import getSampleStyleSheet
    doc = SimpleDocTemplate("svgtest.pdf")
    styles = getSampleStyleSheet()
    style= styles["Normal"]
    Story = [Paragraph("Before the image",style),
             SVGImage(sys.argv[1]),
             Paragraph("After the image",style)]
    doc.build(Story)
    


# -*- coding: utf-8 -*-

# Singleton config object

import ConfigParser
import os
from simplejson import dumps, loads

cfdir=os.path.join(os.path.expanduser('~'),'.rst2pdf')
cfname=os.path.join(cfdir,'config')

def getValue(section,key,default=None):
  section=section.lower()
  key=key.lower()
  try:
    return loads(conf.get (section,key))
  except:
    return default

def setValue(section,key,value):
  section=str(section)
  key=str(key)
  section=section.lower()
  key=key.lower()
  value=dumps(value)
  try:
    r=conf.set(section,key,value)
  except ConfigParser.NoSectionError:
    conf.add_section(section)
    r=conf.set(section,key,value)
  f=open(cfname,'w')
  conf.write(f)
  return r


class ConfigError(Exception):
  def __init__(self,modulename,msg):
    self.modulename=modulename
    self.msg=msg


conf=ConfigParser.SafeConfigParser()
if not os.path.isdir(cfdir):
  os.mkdir(cfdir)

if not os.path.isfile(cfname):
  open(cfname, 'w').close()
f=open(cfname,'r')
conf.readfp(f)
f.close()



#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.8/rst2pdf/utils.py $
#$LastChangedDate: 2008-09-03 17:36:52 -0300 (Wed, 03 Sep 2008) $
#$LastChangedRevision: 265 $
import sys
from reportlab.platypus import PageBreak, Spacer
from flowables import *
import shlex
from log import log

def parseRaw (data):
    '''Parse and process a simple DSL to handle creation of flowables.

    Supported (can add others on request):

    * PageBreak

    * Spacer width, height

    '''
    elements=[]
    lines=data.splitlines()
    for line in lines:
        lexer=shlex.shlex(line)
        lexer.whitespace+=','
        tokens=list(lexer)
        command=tokens[0]
        if command == 'PageBreak':
            if len(tokens)==1:
                elements.append(MyPageBreak())
            else:
                elements.append(MyPageBreak(tokens[1]))
        if command == 'Spacer':
            elements.append(Spacer(int(tokens[1]),int(tokens[2])))
    return elements

# Looks like this is not used anywhere now
#def depth (node):
#    if node.parent==None:
#        return 0
#    else:
#        return 1+depth(node.parent)



# -*- coding: utf-8 -*-
#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.8/rst2pdf/pygments_code_block_directive.py $
#$LastChangedDate: 2008-09-03 17:36:52 -0300 (Wed, 03 Sep 2008) $
#$LastChangedRevision: 265 $

# :Author: a Pygments author|contributor; Felix Wiemann; Guenter Milde
# :Date: $Date: 2008-09-03 17:36:52 -0300 (Wed, 03 Sep 2008) $
# :Copyright: This module has been placed in the public domain.
# 
# This is a merge of `Using Pygments in ReST documents`_ from the pygments_
# documentation, and a `proof of concept`_ by Felix Wiemann.
# 
# ========== ===========================================================
# 2007-06-01 Removed redundancy from class values.
# 2007-06-04 Merge of successive tokens of same type
#            (code taken from pygments.formatters.others).
# 2007-06-05 Separate docutils formatter script
#            Use pygments' CSS class names (like the html formatter)
#            allowing the use of pygments-produced style sheets.
# 2007-06-07 Merge in the formatting of the parsed tokens
#            (misnamed as docutils_formatter) as class DocutilsInterface
# 2007-06-08 Failsave implementation (fallback to a standard literal block 
#            if pygments not found)
# ========== ===========================================================
# 
# ::

"""Define and register a code-block directive using pygments
"""

# Requirements
# ------------
# ::

from docutils import nodes
from docutils.parsers.rst import directives
try:
    import pygments
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters.html import _get_ttype_class
except ImportError:
    pass

import codecs
from log import log

# Customisation
# -------------
# 
# Do not insert inline nodes for the following tokens.
# (You could add e.g. Token.Punctuation like ``['', 'p']``.) ::

unstyled_tokens = ['']

# DocutilsInterface
# -----------------
# 
# This interface class combines code from
# pygments.formatters.html and pygments.formatters.others.
# 
# It does not require anything of docutils and could also become a part of
# pygments::

class DocutilsInterface(object):
    """Parse `code` string and yield "classified" tokens.
    
    Arguments
    
      code     -- string of source code to parse
      language -- formal language the code is written in.
    
    Merge subsequent tokens of the same token-type. 
    
    Yields the tokens as ``(ttype_class, value)`` tuples, 
    where ttype_class is taken from pygments.token.STANDARD_TYPES and 
    corresponds to the class argument used in pygments html output.

    """

    def __init__(self, code, language):
        self.code = code
        self.language = language
        
    def lex(self):
        # Get lexer for language (use text as fallback)
        try:
            lexer = get_lexer_by_name(self.language)
        except ValueError:
            log.info( "no pygments lexer for %s, using 'text'" % self.language)
            # what happens if pygment isn't present ?
            lexer = get_lexer_by_name('text')
        return pygments.lex(self.code, lexer)
        
            
    def join(self, tokens):
        """join subsequent tokens of same token-type
        """
        tokens = iter(tokens)
        (lasttype, lastval) = tokens.next()
        for ttype, value in tokens:
            if ttype is lasttype:
                lastval += value
            else:
                yield(lasttype, lastval)
                (lasttype, lastval) = (ttype, value)
        yield(lasttype, lastval)

    def __iter__(self):
        """parse code string and yield "clasified" tokens
        """
        try:
            tokens = self.lex()
        except IOError:
            log.info("Pygments lexer not found, using fallback")
            # TODO: write message to INFO 
            yield ('', self.code)
            return

        for ttype, value in self.join(tokens):
            yield (_get_ttype_class(ttype), value)



# code_block_directive
# --------------------
# ::

def code_block_directive(name, arguments, options, content, lineno,
                       content_offset, block_text, state, state_machine):
    """parse and classify content of a code_block
    """
    if 'include' in options:
        try:
            content=codecs.open(options['include'],'r','utf-8').read()
        except IOError, UnicodeError: # no file or problem finding it or reading it
            log.error('Error reading file: "%s" L %s' % (options['include'],lineno))
            content=u''
        if content:
            # here we define the start-at and end-at options 
            # so that limit is included in extraction
            # this is different than the start-after directive of docutils
            # (docutils/parsers/rst/directives/misc.py L73+)
            # which excludes the beginning
            # the reason is we want to be able to define a start-at like
            # def mymethod(self)
            # and have such a definition included
            #import pdb; pdb.set_trace()
            after_text = options.get('start-at', None)
            if after_text:
                # skip content in include_text before *and NOT incl.* a matching text
                after_index = content.find(after_text)
                if after_index < 0:
                    raise state_machine.reporter.severe('Problem with "start-at" option of "%s" '
                                      'code-block directive:\nText not found.' % options['start-at'])
                content = content[after_index:]
                
            after_text = options.get('start-after', None)
            if after_text:
                # skip content in include_text before *and incl.* a matching text
                print type(content),type(after_text)
                after_index = content.find(after_text)
                if after_index < 0:
                    raise state_machine.reporter.severe('Problem with "start-after" option of "%s" '
                                      'code-block directive:\nText not found.' % options['start-after'])
                content = content[after_index + len(after_text):]
                
                
            # same changes here for the same reason
            before_text = options.get('end-at', None)
            if before_text:
                # skip content in include_text after *and incl.* a matching text
                before_index = content.find(before_text)
                if before_index < 0:
                    raise state_machine.reporter.severe('Problem with "end-at" option of "%s" '
                                      'code-block directive:\nText not found.' % options['end-at'])
                content = content[:before_index + len(before_text)]
                
            before_text = options.get('end-before', None)
            if before_text:
                # skip content in include_text after *and NOT incl.* a matching text
                before_index = content.find(before_text)
                if before_index < 0:
                    raise state_machine.reporter.severe('Problem with "end-before" option of "%s" '
                                      'code-block directive:\nText not found.' % options['end-before'])
                content = content[:before_index]

    else:
        content=u'\n'.join(content)

    language = arguments[0]
    # create a literal block element and set class argument
    code_block = nodes.literal_block(classes=["code", language])
    
    # parse content with pygments and add to code_block element
    for cls, value in DocutilsInterface(content, language):
        if cls in unstyled_tokens:
            # insert as Text to decrease the verbosity of the output.
            code_block += nodes.Text(value, value)
        else:
            code_block += nodes.inline(value, value, classes=["pygments-"+cls])

    return [code_block]


# Register Directive
# ------------------
# ::

code_block_directive.arguments = (1, 0, 1)
code_block_directive.content = 1
code_block_directive.options = { 'include' : directives.unchanged_required,
                                 'start-at' : directives.unchanged_required,
                                 'end-at' : directives.unchanged_required,
                                 'start-after' : directives.unchanged_required,
                                 'end-before' : directives.unchanged_required,
                                 }

directives.register_directive('code-block', code_block_directive)

# .. _doctutils: http://docutils.sf.net/
# .. _pygments: http://pygments.org/
# .. _Using Pygments in ReST documents: http://pygments.org/docs/rstdirective/
# .. _proof of concept:
#      http://article.gmane.org/gmane.text.docutils.user/3689
# 
# Test output
# -----------
# 
# If called from the command line, call the docutils publisher to render the
# input::

if __name__ == '__main__':
    from docutils.core import publish_cmdline, default_description
    description = "code-block directive test output" + default_description
    try:
        import locale
        locale.setlocale(locale.LC_ALL, '')
    except:
        pass
    publish_cmdline(writer_name='pdf', description=description)
    





# -*- coding: utf-8 -*-
#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.8/rst2pdf/createpdf.py $
#$LastChangedDate: 2008-09-12 13:26:09 -0300 (Fri, 12 Sep 2008) $
#$LastChangedRevision: 353 $
__docformat__ = 'reStructuredText'

import re
import sys
import os
from os.path import join, abspath, dirname
import pprint
import string
from types import StringType
from urlparse import *
from copy import copy
from cgi import escape
import logging
from optparse import OptionParser
from docutils import __version__, __version_details__, SettingsSpec
from docutils import frontend, io, utils, readers, writers
from docutils.transforms import Transformer
import docutils.readers.doctree
import docutils.core
import docutils.nodes

import pygments_code_block_directive

import reportlab
from reportlab.platypus import *
from reportlab.pdfbase.pdfmetrics import stringWidth
import reportlab.lib.colors as colors
from reportlab.lib.enums import *
from reportlab.lib.units import *
from reportlab.lib.pagesizes import *

from flowables import *
from svgimage import SVGImage
from log import log

import config

#def escape (x,y):
#    "Dummy escape function to test for excessive escaping"
#    return x
from utils import log,parseRaw
import styles as sty

try:
    from PIL import Image as PILImage
except ImportError:
    try:
        import Image as PILImage
    except ImportError:
        log.warning("No support for images other than JPG, and limited support for image size. Please install PIL")

try:
    import wordaxe
    from wordaxe.PyHnjHyphenator import PyHnjHyphenator
    from wordaxe.rl.paragraph import Paragraph
    from wordaxe.rl.styles import ParagraphStyle, getSampleStyleSheet
    haveWordaxe=True
except ImportError:
    #log.warning("No support for hyphenation, install wordaxe")
    haveWordaxe=False

try:
    import sphinx
    HAS_SPHINX=True
except ImportError:
    HAS_SPHINX=False

class RstToPdf(object):

    def __init__(self, stylesheets = [], language = 'en_US',
                 breaklevel=1,fontPath=[],fitMode='shrink',sphinx=False):
        global HAS_SPHINX
        self.lowerroman=['i','ii','iii','iv','v','vi','vii','viii','ix','x','xi']
        self.loweralpha=string.ascii_lowercase
        self.doc_title=None
        self.doc_author=None
        self.decoration = {'header':None, 'footer':None, 'endnotes':[]}
        stylesheets = [os.path.join(abspath(dirname(__file__)),'styles','styles.json')]+stylesheets
        self.styles=sty.StyleSheet(stylesheets,fontPath)
        self.breaklevel=breaklevel
        self.fitMode=fitMode

        # Sorry about this, but importing sphinx.roles makes some
        # ordinary documents fail (demo.txt specifically) so
        # I can' t just try to import it outside. I need
        # to do it only if it's requested
        if HAS_SPHINX and sphinx:
            import sphinx.roles
        else:
            HAS_SPHINX=False

        # Load the hyphenators for all required languages
        if haveWordaxe:
            if not self.styles.languages:
                self.styles.languages=[language]
                self.styles['bodytext'].language=language
            for lang in self.styles.languages:
                try:
                    wordaxe.hyphRegistry[lang] = PyHnjHyphenator(lang,5)
                except ImportError: #PyHnj C extension is not installed
                    wordaxe.hyphRegistry[lang] = PyHnjHyphenator(lang,5,purePython=True)
            log.info('hyphenation by default in %s , loaded %s',
                self.styles['bodytext'].language, ','.join(self.styles.languages))

    def styleToFont(self, style):
        '''Takes a style name, returns a font tag for it, like
        "<font face=helvetica size=14 color=red>". Used for inline
        nodes (custom interpreted roles)'''

        try:
            s=self.styles[style]
            bc=s.backColor
            if bc:
                r='<font face="%s" size="%d" color="#%s" backColor="#%s">' % (s.fontName,s.fontSize,
                                                                              s.textColor.hexval()[2:],
                                                                              bc.hexval()[2:])
            else:
                r='<font face="%s" size="%d" color="#%s">' % (s.fontName,s.fontSize,s.textColor.hexval()[2:])
            return r
        except KeyError:
            log.warning('Unknown class %s', style)
            return None


    def gather_pdftext (self, node, depth,replaceEnt=True):
        return ''.join([self.gen_pdftext(n,depth,replaceEnt) for n in node.children ])

    def gen_pdftext(self, node, depth, replaceEnt=True):
        pre=""
        post=""

        log.debug("self.gen_pdftext: %s", node.__class__)
        try:
            log.debug("self.gen_pdftext: %s", node)
        except (UnicodeDecodeError, UnicodeEncodeError):
            log.debug("self.gen_pdftext: %r", node)

        if isinstance (node, docutils.nodes.paragraph) \
            or isinstance (node, docutils.nodes.title) \
            or isinstance (node, docutils.nodes.subtitle):
            node.pdftext=self.gather_pdftext(node,depth)+"\n"

        elif isinstance (node, docutils.nodes.Text):
            node.pdftext=node.astext()
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.strong):
            pre="<b>"
            post="</b>"
            node.pdftext=self.gather_pdftext(node,depth)
            #if replaceEnt:
            #    node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.emphasis):
            pre="<i>"
            post="</i>"
            node.pdftext=self.gather_pdftext(node,depth)
            #if replaceEnt:
            #    node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.literal):
            pre='<font face="%s">'%self.styles['code'].fontName
            post="</font>"
            node.pdftext=self.gather_pdftext(node,depth)
            #if replaceEnt:
            #    node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.superscript):
            pre='<super>'
            post="</super>"
            node.pdftext=self.gather_pdftext(node,depth)
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.subscript):
            pre='<sub>'
            post="</sub>"
            node.pdftext=self.gather_pdftext(node,depth)
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.title_reference):
            pre=self.styleToFont("title_reference")
            post="</font>"
            node.pdftext=self.gather_pdftext(node,depth)
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.reference) :
            uri=node.get('refuri')
            if uri:
                if urlparse(uri)[0]:
                    pre+=u'<a href="%s" color="%s">'%(uri,self.styles.linkColor)
                    post='</a>'+post
            else:
                uri=node.get('refid')
                if uri:
                    pre+=u'<a href="#%s" color="%s">'%(uri,self.styles.linkColor)
                    post='</a>'+post
            node.pdftext=self.gather_pdftext(node,depth)
            #if replaceEnt:
            #    node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.option_string) \
            or isinstance (node, docutils.nodes.option_argument):
            node.pdftext=node.astext()
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)

        elif isinstance (node, docutils.nodes.header) \
            or isinstance (node, docutils.nodes.footer):
            node.pdftext=self.gather_pdftext(node,depth)
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)

            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.system_message)     \
            or isinstance (node, docutils.nodes.problematic):
            pre='<font color="red">'
            post="</font>"
            node.pdftext=self.gather_pdftext(node,depth)
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.generated):
            node.pdftext=self.gather_pdftext(node,depth)
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext+post

        elif isinstance (node, docutils.nodes.image):
            node.pdftext='<img src="%s" />'%node.get('uri')

        elif isinstance (node, docutils.nodes.footnote_reference):
            # Fixme link to the right place
            anchors=''.join(['<a name="%s"/>'%i for i in node['ids'] ])
            node.pdftext=u'%s<super><a href="%s" color="%s">%s</a></super>'%(
                anchors,'#'+node.astext(),self.styles.linkColor,node.astext())
        elif isinstance (node, docutils.nodes.citation_reference):
            # Fixme link to the right place
            anchors=''.join(['<a name="%s"/>'%i for i in node['ids'] ])
            node.pdftext=u'%s[<a href="%s" color="%s">%s</a>]'%(
                anchors,'#'+node.astext(),self.styles.linkColor,node.astext())

        elif isinstance (node, docutils.nodes.target):
            pre=u'<a name="%s"/>'%node['ids'][0]
            node.pdftext=self.gather_pdftext(node,depth)
            if replaceEnt:
                node.pdftext=escape(node.pdftext,True)
            node.pdftext=pre+node.pdftext

        elif isinstance (node, docutils.nodes.inline):
            ftag=self.styleToFont(node['classes'][0])
            if ftag:
                node.pdftext="%s%s</font>"%(ftag,self.gather_pdftext(node,depth))
            else:
                node.pdftext=self.gather_pdftext(node,depth)

        else:
            log.warning("Unkn. node (self.gen_pdftext): %s", node.__class__)
            try:
                log.debug(node)
            except (UnicodeDecodeError, UnicodeEncodeError):
                log.debug(repr(node))
            node.pdftext=self.gather_pdftext(node,depth)
            #print node.transform

        try:
            log.debug("self.gen_pdftext: %s" % node.pdftext)
        except UnicodeDecodeError:
            pass
        return node.pdftext

    def gen_elements(self, node, depth, style=None):

        log.debug("gen_elements: %s", node.__class__)
        try:
            log.debug("gen_elements: %s", node)
        except (UnicodeDecodeError, UnicodeEncodeError):
            log.debug("gen_elements: %r", node)

        if style is None:
            style=self.styles.styleForNode(node)

        try:
            if node['classes']:
                # Supports only one class, sorry ;-)
                try:
                    style=self.styles[node['classes'][0]]
                except:
                    log.info("Unknown class %s, using class bodytext.", node['classes'][0])
        except TypeError: # Happens when a docutils.node.Text reaches here
            pass

        if isinstance (node, docutils.nodes.document):
            node.elements=self.gather_elements(node,depth,style=style)

        elif HAS_SPHINX and isinstance(node,sphinx.addnodes.glossary):
            node.elements=self.gather_elements(node,depth,style=style)
          
        #######################
        ## Tables
        #######################

        elif isinstance (node, docutils.nodes.table):
            # FIXME: make spacing configurable
            node.elements=[Spacer(0,6)]+self.gather_elements(node,depth)

        elif isinstance (node, docutils.nodes.tgroup):
            rows=[]
            colwidths=[]
            hasHead=False
            headRows=0
            for n in node.children:
                if isinstance (n,docutils.nodes.thead):
                    hasHead=True
                    for row in n.children:
                        r=[]
                        for cell in row.children:
                            r.append(cell)
                        rows.append(r)
                    headRows=len(rows)
                elif isinstance (n,docutils.nodes.tbody):
                    for row in n.children:
                        r=[]
                        for cell in row.children:
                            r.append(cell)
                        rows.append(r)
                elif isinstance (n,docutils.nodes.colspec):
                    colwidths.append(int(n['colwidth']))

            spans=self.filltable (rows)

            data=[]
            cellStyles=[]
            rowids=range(0,len(rows))
            for row,i in zip(rows,rowids):
                r=[]
                j=0
                for cell in row:
                    if isinstance(cell,str):
                        r.append("")
                    else:
                        if i<headRows:
                            ell=self.gather_elements(cell,depth,style=self.styles['table-heading'])
                        else:
                            ell=self.gather_elements(cell,depth,style=style)
                        if len(ell)==1:
                            # Experiment: if the cell has a single element, extract its
                            # class and use it for the cell. That way, you can have cells
                            # with specific background colors, at least
                            cellStyles+=self.styles.pStyleToTStyle(ell[0].style,j,i)
                        r.append(ell)
                    j+=1
                data.append(r)

            st=spans+sty.tstyleNorm+self.styles.tstyleBody()+cellStyles

            if hasHead:
                st+=self.styles.tstyleHead(headRows)

            #node.elements=[Table(data,colWidths=colwidths,style=TableStyle(st))]
            node.elements=[DelayedTable(data,colwidths,st)]

        elif isinstance (node, docutils.nodes.title):
            # Special cases: (Not sure this is right ;-)
            if isinstance (node.parent, docutils.nodes.document):
                # FIXME maybe make it a coverpage?
                node.elements=[Paragraph(self.gen_pdftext(node,depth), self.styles['title'])]
                self.doc_title=unicode(self.gen_pdftext(node,depth)).strip()
            elif isinstance (node.parent, docutils.nodes.topic):
                # FIXME style correctly
                node.elements=[Paragraph(self.gen_pdftext(node,depth), self.styles['heading3'])]
            elif isinstance (node.parent, docutils.nodes.admonition):
                node.elements=[Paragraph(self.gen_pdftext(node,depth), self.styles['heading3'])]
            elif isinstance (node.parent, docutils.nodes.table):
                node.elements=[Paragraph(self.gen_pdftext(node,depth), self.styles['table-title'])]
            elif isinstance (node.parent, docutils.nodes.sidebar):
                node.elements=[Paragraph(self.gen_pdftext(node,depth), self.styles['sidebar-title'])]
            else:
                # Section/Subsection/etc.
                text=self.gen_pdftext(node,depth)
                fch=node.children[0]
                if isinstance(fch,docutils.nodes.generated) and fch['classes']==['sectnum']:
                    snum=fch.astext()
                else:
                    snum=None
                key=node.get('refid')
                node.elements=[OutlineEntry(key,text,depth-1,snum),
                               Paragraph(text, self.styles['heading%d' % min(depth,3)])]
                if depth <=self.breaklevel:
                    node.elements.insert(0,MyPageBreak())


        elif isinstance (node, docutils.nodes.subtitle):
            if isinstance (node.parent,docutils.nodes.sidebar):
                node.elements=[Paragraph(self.gen_pdftext(node,depth), self.styles['sidebar-subtitle'])]
            elif isinstance (node.parent,docutils.nodes.document):
                node.elements=[Paragraph(self.gen_pdftext(node,depth), self.styles['subtitle'])]

        elif HAS_SPHINX and isinstance(node,sphinx.addnodes.compact_paragraph):
            node.elements=self.gather_elements(node,depth,style=style)

        elif isinstance (node, docutils.nodes.paragraph):
            node.elements=[Paragraph(self.gen_pdftext(node,depth), style)]

        elif isinstance (node, docutils.nodes.docinfo):
            # A docinfo usually contains several fields.
            # We'll render it as a series of elements, one field each.

            node.elements=self.gather_elements(node,depth,style=style)

        elif isinstance (node, docutils.nodes.field):
            # A field has two child elements, a field_name and a field_body.
            # We render as a two-column table, left-column is right-aligned,
            # bold, and much smaller

            fn=Paragraph(self.gather_pdftext(node.children[0],depth)+":",style=self.styles['fieldname'])
            fb=self.gen_elements(node.children[1],depth)
            node.elements=[Table([[fn,fb]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])]

        elif isinstance (node, docutils.nodes.decoration):
            # This is a tricky one. We need to switch our document's
            # page templates based on this. If decoration contains a
            # header and/or a footer, we need to use those
            # right now, we avoid trouble.
            # FIXME Implement
            node.elements=self.gather_elements(node,depth,style=style)

        elif isinstance (node, docutils.nodes.header):
            self.decoration['header']=self.gather_pdftext(node,depth)
            node.elements=[]
        elif isinstance (node, docutils.nodes.footer):
            self.decoration['footer']=self.gather_pdftext(node,depth)
            node.elements=[]

        elif isinstance (node, docutils.nodes.author):
            if isinstance (node.parent,docutils.nodes.authors):
                # Is only one of multiple authors. Return a paragraph
                node.elements=[Paragraph(self.gather_pdftext(node,depth), style=style)]
            else:
                # A single author: works like a field
                fb=self.gather_pdftext(node,depth)
                node.elements=[Table([[Paragraph("Author:",style=self.styles['fieldname']),
                                    Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])]
                self.doc_author=fb.strip()

        elif isinstance (node, docutils.nodes.authors):
            # Multiple authors. Create a two-column table. Author references on the right.
            td=[[Paragraph("Authors:",style=self.styles['fieldname']),self.gather_elements(node,depth,style=style)]]
            node.elements=[Table(td,style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])]

        elif isinstance (node, docutils.nodes.organization):
            fb=self.gather_pdftext(node,depth)
            t=Table([[Paragraph("Organization:",style=self.styles['fieldname']),
                                Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])
            node.elements=[t]
        elif isinstance (node, docutils.nodes.contact):
            fb=self.gather_pdftext(node,depth)
            t=Table([[ Paragraph("Contact:",style=self.styles['fieldname']),
                                Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])
            node.elements=[t]
        elif isinstance (node, docutils.nodes.address):
            fb=self.gather_pdftext(node,depth)
            t=Table([[ Paragraph("Address:",style=self.styles['fieldname']),
                                Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])
            node.elements=[t]
        elif isinstance (node, docutils.nodes.version):
            fb=self.gather_pdftext(node,depth)
            t=Table([[ Paragraph("Version:",style=self.styles['fieldname']),
                                Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])
            node.elements=[t]
        elif isinstance (node, docutils.nodes.revision):
            fb=self.gather_pdftext(node,depth)
            t=Table([[ Paragraph("Revision:",style=self.styles['fieldname']),
                                Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])
            node.elements=[t]
        elif isinstance (node, docutils.nodes.status):
            fb=self.gather_pdftext(node,depth)
            t=Table([[ Paragraph("Status:",style=self.styles['fieldname']),
                                Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])
            node.elements=[t]
        elif isinstance (node, docutils.nodes.date):
            fb=self.gather_pdftext(node,depth)
            t=Table([[ Paragraph("Date:",style=self.styles['fieldname']),
                                Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])
            node.elements=[t]
        elif isinstance (node, docutils.nodes.copyright):
            fb=self.gather_pdftext(node,depth)
            t=Table([[Paragraph("Copyright:",style=self.styles['fieldname']),
                                Paragraph(fb,style) ]],style=sty.tstyles['field'],colWidths=[sty.fieldlist_lwidth,None])
            node.elements=[t]

        elif isinstance (node, docutils.nodes.topic)                                \
            or isinstance (node, docutils.nodes.field_body):
            node.elements=self.gather_elements(node,depth,style=style)

        elif isinstance (node, docutils.nodes.section):
            node.elements=self.gather_elements(node,depth+1)

        elif isinstance (node, docutils.nodes.bullet_list)                   \
            or isinstance (node, docutils.nodes.enumerated_list)            \
            or isinstance (node, docutils.nodes.definition_list)            \
            or isinstance (node, docutils.nodes.option_list)                \
            or isinstance (node, docutils.nodes.field_list)                 \
            or isinstance (node, docutils.nodes.definition):

            node.elements=self.gather_elements(node,depth,style=style)

        elif isinstance (node, docutils.nodes.option_list_item):

            optext = ', '.join([self.gather_pdftext(child,depth) for child in node.children[0].children])
            desc = self.gather_elements(node.children[1],depth,style)

            node.elements=[Table([[self.PreformattedFit(optext,self.styles["literal"]),desc]],style=sty.tstyles['field'])]


        elif isinstance (node, docutils.nodes.definition_list_item):

            # I need to catch the classifiers here
            tt=[]
            dt=[]
            for n in node.children:
                if isinstance(n,docutils.nodes.term):
                    for i in n['ids']: # Used by sphinx glossary lists
                        tt.append('<a name="%s"/>'%i)
                    tt.append(self.styleToFont("definition_list_term")+self.gather_pdftext(n,depth,style)+"</font>")
                elif isinstance(n,docutils.nodes.classifier) :
                    tt.append(self.styleToFont("definition_list_classifier")+self.gather_pdftext(n,depth,style)+"</font>")
                else:
                    dt=dt+self.gen_elements(n,depth,style)

            node.elements=[Paragraph(' : '.join(tt),self.styles['definition_list_term']),
                           MyIndenter(left=10)]+dt+[MyIndenter(left=-10)]

        elif isinstance (node, docutils.nodes.list_item):
            # A list_item is a table of two columns.
            # The left one is the bullet itself, the right is the
            # item content. This way we can nest them.

            el=self.gather_elements(node,depth,style=style)
            b=""
            if node.parent.get('bullet') or isinstance(node.parent,docutils.nodes.bullet_list):
                b=node.parent.get('bullet')
                if b=="None":
                    b=""

            elif node.parent.get ('enumtype')=='arabic':
                b=str(node.parent.children.index(node)+1)+'.'

            elif node.parent.get ('enumtype')=='lowerroman':
                b=str(self.lowerroman[node.parent.children.index(node)])+'.'

            elif node.parent.get ('enumtype')=='upperroman':
                b=str(self.lowerroman[node.parent.children.index(node)].upper())+'.'

            elif node.parent.get ('enumtype')=='loweralpha':
                b=str(self.loweralpha[node.parent.children.index(node)])+'.'
            elif node.parent.get ('enumtype')=='upperalpha':
                b=str(self.loweralpha[node.parent.children.index(node)].upper())+'.'
            else:
                log.critical("Unknown kind of list_item %s", node.parent)
                sys.exit(1)
            # FIXME: use different unicode bullets depending on b
            if b and b in "*+-":
                b=u'\u2022'

            el[0].bulletText = b
            for e in el[1:]:
                e.bulletText=" "
            node.elements=[Table([[el]],style=sty.tstyles["bullet"])]

        elif isinstance (node, docutils.nodes.transition):
            node.elements=[Separation()]


        elif isinstance (node, docutils.nodes.system_message)     \
            or isinstance (node, docutils.nodes.problematic):
            # FIXME show the error in the document, red, whatever
            log.warning("Problematic node %s", node.astext())
            node.elements=[]

        elif isinstance (node, docutils.nodes.block_quote):
            node.elements=[MyIndenter(left=20)]+self.gather_elements(node,depth,style)+[MyIndenter(left=-20)]

        elif isinstance (node, docutils.nodes.attribution):
            node.elements=[Paragraph(self.gather_pdftext(node,depth),self.styles['attribution'])]

        elif isinstance (node, docutils.nodes.comment):
            # Class that generates no output
            node.elements=[]

        elif isinstance (node, docutils.nodes.line_block):
            # Obsolete? Let's do something anyway.
            # FIXME: indent or not?
            qstyle=copy(style)
            qstyle.leftIndent+=30
            node.elements=self.gather_elements(node,depth,style=qstyle)

        elif isinstance (node, docutils.nodes.line):
            # All elements in one line
            node.elements=[Paragraph(self.gather_pdftext(node,depth),style=style)]

        elif isinstance (node, docutils.nodes.literal_block) \
            or isinstance (node, docutils.nodes.doctest_block):
            node.elements=[self.PreformattedFit(self.gather_pdftext(node,depth,replaceEnt=True),self.styles['code'])]

        elif isinstance (node, docutils.nodes.attention)        \
            or isinstance (node, docutils.nodes.caution)            \
            or isinstance (node, docutils.nodes.danger)             \
            or isinstance (node, docutils.nodes.error)                \
            or isinstance (node, docutils.nodes.hint)                 \
            or isinstance (node, docutils.nodes.important)        \
            or isinstance (node, docutils.nodes.note)                 \
            or isinstance (node, docutils.nodes.tip)                    \
            or isinstance (node, docutils.nodes.warning)            \
            or isinstance (node, docutils.nodes.admonition):
            node.elements=[Paragraph(node.tagname.title(),style=self.styles['heading3'])]+self.gather_elements(node,depth,style=style)

        elif isinstance (node, docutils.nodes.image):
            # FIXME: handle all the other attributes

            imgname=str(node.get("uri"))

            # Figuring out the size to display of an image is ... annoying.
            # If the user provides a size with a unit, it's simple, adjustUnits
            # will return it in points and we're done.

            # However, often the unit wil be "%" (specially if it's meant for
            # HTML originally.In which case, we will use the following:

            # Find the image size in pixels:

            try:
                # FIXME find extensions uniconvertor supports
                if imgname.split('.')[-1].lower() in ["ai","ccx","cdr","cgm","cmx","sk1","sk","svg","xml","wmf","fig"]:
                    iw,ih=SVGImage(imgname).wrap(0,0)
                else:
                    img=PILImage.open(imgname)
                    iw,ih=img.size
                # Assume a DPI of 300, which is pretty much made up,
                # and then a 100% size would be iw*inch/300, so we pass
                # that as the second parameter to adjustUnits

                w=node.get('width')
                if w is not None:
                    if iw:
                        w=self.styles.adjustUnits(w,iw*inch/300)
                    else:
                        w=self.styles.adjustUnits(w,self.styles.pw*.5)
                else:
                    log.warning("Using image %s without specifying size."
                        "Calculating based on 300dpi", imgname)
                    # No width specified at all. Make it up
                    # as if we knew what we're doing
                    if iw:
                        w=iw*inch/300
                    else:
                        w=2*inch # We are getting desperate here ;-)

                h=node.get('height')
                if h is not None:
                    if ih:
                        h=self.styles.adjustUnits(h,ih*inch/300)
                    else:
                        h=self.styles.adjustUnits(h,self.styles.ph*.5)
                else:
                    # Now, often, only the width is specified!
                    # if we don't have a height, we need to keep the
                    # aspect ratio, or else it will look ugly
                    if iw:
                        h=w*ih/iw
                    else:
                        h=2*inch # We are getting desperate here ;-)

                # And now we have this probably completely bogus size!
                log.info("Image %s size calculated:  %fcm by %fcm",
                    imgname, w/cm, h/cm)
                if imgname.split('.')[-1].lower() in ["ai","ccx","cdr","cgm","cmx","sk1","sk","svg","xml","wmf","fig"]:
                    node.elements=[SVGImage(filename=imgname,
                                            height=h,
                                            width=w)]
                else:
                    node.elements=[Image(filename=imgname,
                                         height=h,
                                         width=w)]
                i=node.elements[0]
                if node.get('align'):
                    i.hAlign=node.get('align').upper()
                else:
                    i.hAlign='CENTER'
            except IOError,e: #No image, or no permissions
                log.error('Error opening "%s": %s'%(imgname,str(e)))
                node.elements=[]
                

        elif isinstance (node, docutils.nodes.figure):
            # The sub-elements are the figure and the caption, and't ugly if
            # they separate
            #node.elements=[KeepTogether(self.gather_elements(node,depth,style=self.styles["figure"]))]
            # FIXME: using KeepTogether as a container fails inside a float.
            node.elements=self.gather_elements(node,depth,style=self.styles["figure"])

        elif isinstance (node, docutils.nodes.caption):
            node.elements=[Paragraph('<i>'+self.gather_pdftext(node,depth)+'</i>',style=style)]

        elif isinstance (node, docutils.nodes.legend):
            node.elements=self.gather_elements(node,depth,style=style)

        elif isinstance (node, docutils.nodes.sidebar):
            node.elements=self.gather_elements(node,depth,style=None)
            style=self.styles['sidebar']

        elif isinstance (node, docutils.nodes.rubric):
            node.elements=[Paragraph(self.gather_pdftext(node,depth),self.styles['rubric'])]

        elif isinstance (node, docutils.nodes.compound):
            # FIXME think if this is even implementable
            node.elements=self.gather_elements(node,depth,style)

        elif isinstance (node, docutils.nodes.container):
            # FIXME think if this is even implementable
            node.elements=self.gather_elements(node,depth,style)

        elif isinstance (node, docutils.nodes.substitution_definition):
            node.elements=[]

        elif isinstance (node, docutils.nodes.tbody):
            rows=[self.gen_elements(n,depth) for n in node.children]
            t=[]
            for r in rows:
                if not r:
                    continue
                t.append(r)
            node.elements=[Table(t,style=sty.tstyles['normal'])]

        elif isinstance (node, docutils.nodes.footnote) or \
            isinstance (node, docutils.nodes.citation):
            # It seems a footnote contains a label and a series of elements
            ltext=self.gather_pdftext(node.children[0],depth)
            if len(node['backrefs'])>1:
                backrefs=[]
                i=1
                for r in node['backrefs']:
                    backrefs.append('<a href="#%s" color="%s">%d</a>'%(
                        r,self.styles.linkColor,i))
                    i+=1
                backrefs='(%s)'%', '.join(backrefs)
                label=Paragraph('<a name="%s"/>%s'%(ltext,ltext+backrefs),self.styles["normal"])
            elif len(node['backrefs'])==1:
                label=Paragraph('<a name="%s"/><a href="%s" color="%s">%s</a>'%(
                    ltext,node['backrefs'][0],self.styles.linkColor,ltext),self.styles["normal"])
            else:
                label=Paragraph('<a name="%s"/>%s'%(ltext,ltext),self.styles["normal"])
            contents=self.gather_elements(node,depth,style)[1:]
            self.decoration['endnotes'].append([label,contents])
            node.elements=[]

        elif isinstance (node, docutils.nodes.label):
            node.elements=[Paragraph(self.gather_pdftext(node,depth),style)]
        elif isinstance (node, docutils.nodes.Text):
            node.elements=[Paragraph(self.gather_pdftext(node,depth),style)]
        elif isinstance (node, docutils.nodes.entry):
            node.elements=self.gather_elements(node,depth,style)

        elif isinstance (node, docutils.nodes.target):
            node.elements=self.gather_elements(node,depth,style)

        elif isinstance (node, docutils.nodes.reference):
            node.elements=self.gather_elements(node,depth,style)

        elif isinstance (node, docutils.nodes.raw):
            # Not really raw, but what the heck
            node.elements=parseRaw(str(node.astext()))

        # FIXME nodes we are ignoring for the moment
        elif isinstance (node, docutils.nodes.citation):
            node.elements=[]
        else:
            log.error("Unkn. node (gen_elements): %s", str(node.__class__))
            # Why fail? Just log it and do our best.
            try:
                log.debug(node)
            except (UnicodeDecodeError, UnicodeEncodeError):
                log.debug(repr(node))
            node.elements=self.gather_elements(node,depth,style)
            #sys.exit(1)

        # set anchors for internal references
        try:
            for i in node['ids']:
                # ids should link **after** pagebreaks
                if len(node.elements) and isinstance(node.elements[0],MyPageBreak):
                    idx=1
                else:
                    idx=0
                node.elements.insert(idx,Reference(i))
        except TypeError: #Hapens with docutils.node.Text
            pass

        if 'float' in style.__dict__:
            node.elements=[Sidebar(node.elements,style)]
        elif 'width' in style.__dict__:
            node.elements=[BoundByWidth(style.width,node.elements,style,mode="shrink")]
        try:
            log.debug("gen_elements: %s", node.elements)
        except: # unicode problems FIXME: explicit error
            pass
        return node.elements

    def gather_elements (self,node, depth,style=None):
        if style is None:
            style=self.styles.styleForNode(node)
        r=[]
        if 'float' in style.__dict__: style = None # Don't pass floating styles to children!
        for n in node.children:
            #import pdb; pdb.set_trace()
            r.extend(self.gen_elements(n,depth,style=style))
        return r

    def filltable (self,rows):
        """
        Takes a list of rows, consisting of cells and performs the following fixes:

        * For multicolumn cells, add continuation cells, to make all rows the same
        size.

        * For multirow cell, insert continuation cells, to make all columns the
        same size.

        * If there are still shorter rows, add empty cells at the end (ReST quirk)

        * Once the table is *normalized*, create spans list, fitting for reportlab's
        Table class.

        """

        # If there is a multicol cell, we need to insert Continuation Cells
        # to make all rows the same length

        for y in range(0,len(rows)):
            for x in range (0,len(rows[y])):
                cell=rows[y][x]
                if isinstance (cell,str):
                    continue
                if cell.get("morecols"):
                    for i in range(0,cell.get("morecols")):
                        rows[y].insert(x+1,"")

        for y in range(0,len(rows)):
            for x in range (0,len(rows[y])):
                cell=rows[y][x]
                if isinstance (cell,str):
                    continue
                if cell.get("morerows"):
                    for i in range(0,cell.get("morerows")):
                        rows[y+i+1].insert(x,"")


        # If a row is shorter, add empty cells at the right end
        maxw=max([len(r) for r in rows])
        for r in rows:
            while len(r) < maxw:
                r.append("")

        # Create spans list for reportlab's table style
        spans=[]
        for y in range(0,len(rows)):
            for x in range (0,len(rows[y])):
                cell=rows[y][x]
                if isinstance (cell,str):
                    continue
                if cell.get("morecols"):
                    mc=cell.get("morecols")
                else: mc=0
                if cell.get("morerows"):
                    mr=cell.get("morerows")
                else: mr=0

                if mc or mr:
                    spans.append(('SPAN',(x,y),(x+mc,y+mr)))
        return spans

    def PreformattedFit(self,text,style):
        """Preformatted section that gets horizontally compressed if needed."""
        # Pass a ridiculous size, then it will shrink to what's available
        # in the frame
        return BoundByWidth(2000*cm,content=[XPreformatted(text,style)],mode=self.fitMode,style=style)

    def createPdf(self,text=None,output=None,doctree=None,compressed=False):
        '''Create a PDF from text (ReST input), or doctree (docutil nodes)
        and save it in outfile.

        If outfile is a string, it's a filename.
        If it's something with a write method, (like a StringIO,
        or a file object), the data is saved there.

        '''

        if doctree is None:
            if text is not None:
                doctree=docutils.core.publish_doctree(text)
                log.debug(doctree)
            else:
                log.error('Error: createPdf needs a text or a doctree to be useful')
                return
        elements=self.gen_elements(doctree,0)

        # Put the endnotes at the end ;-)
        endnotes = self.decoration['endnotes']
        if endnotes:
            elements.append(Spacer(1,2*cm))
            elements.append(Separation())
            # FIXME: the width of the left column should not be fixed
            for n in self.decoration['endnotes']:
                elements.append(Table([[n[0],n[1]]],
                                style=sty.tstyles['endnote'],
                                colWidths=[sty.endnote_lwidth,None]))

        head=self.decoration['header']
        foot=self.decoration['footer']

        # So, now, create the FancyPage with the right sizes and elements
        FP=FancyPage("fancypage",head,foot,self.styles)

        pdfdoc = FancyDocTemplate(output,
                                 pageTemplates=[FP],
                                 showBoundary=0,
                                 pagesize=self.styles.ps,
                                 title=self.doc_title,
                                 author=self.doc_author,
                                 pageCompression=compressed
                                 )
        pdfdoc.build(elements)

class FancyDocTemplate(BaseDocTemplate):
    pass

class FancyPage(PageTemplate):
    """ A page template that handles changing layouts.
    """
    def __init__(self,_id,head,foot,styles):

        self.styles=styles
        self.head=head
        self.foot=foot
        PageTemplate.__init__(self,_id,[])

    def beforeDrawPage(self,canv,doc):
        '''Do adjustments to the page according to where we are in the document.

           * Gutter margins on left or right as needed

        '''
        
        self.tw=self.styles.pw-self.styles.lm-self.styles.rm-self.styles.gm
        # What page template to use?
        tname=canv.__dict__.get('templateName',self.styles.firstTemplate)
        self.template=self.styles.pageTemplates[tname]
        # Adjust text space accounting for header/footer

        if self.head and self.template.get('showHeader',True):
            self.hh=Paragraph(self.head,style=self.styles['header']).\
                    wrap(self.tw,self.styles.ph)[1]
        else:
            self.hh=0
        if self.head and self.template.get('showFooter',True):
            self.fh=Paragraph(self.foot,style=self.styles['footer']).\
                    wrap(self.tw,self.styles.ph)[1]
        else:
            self.fh=0

        self.hx=self.styles.lm
        self.hy=self.styles.ph-self.styles.tm

        self.fx=self.styles.lm
        self.fy=self.styles.bm
        self.th=self.styles.ph-self.styles.tm-self.styles.bm-self.hh-self.fh


        # Adjust gutter margins
        if doc.page%2: # Left page
            x1=self.styles.lm
            y1=self.styles.tm+self.hh
        else: # Right page
            x1=self.styles.lm+self.styles.gm
            y1=self.styles.tm+self.hh

        self.frames=[]
        for frame in self.template['frames']:
            self.frames.append(SmartFrame(self,self.styles.adjustUnits(frame[0],self.tw)+x1,
                                          self.styles.adjustUnits(frame[1],self.th)+y1,
                                          self.styles.adjustUnits(frame[2],self.tw),
                                          self.styles.adjustUnits(frame[3],self.th)))
        #textframe=Frame(x1,y1,self.tw,self.th)
        #self.frames=[textframe]

    def replaceTokens(self,text,canv,doc):
        ''' Put doc_title/page number/etc in text of header/footer'''

        text=text.replace('###Page###',str(doc.page))
        text=text.replace("###Title###",doc.title)
        # FIXME: make this nicer
        try:
            text=text.replace("###Section###",canv.sectName)
        except AttributeError:
            text=text.replace("###Section###",'')
        try:
            text=text.replace("###SectNum###",canv.sectNum)
        except AttributeError:
            text=text.replace("###SectNum###",'')
        return text

    def afterDrawPage(self,canv,doc):
        '''Draw header/footer'''
        # Adjust for gutter margin
        if doc.page%2: # Left page
            hx=self.hx
            fx=self.fx
        else: # Right Page
            hx=self.hx+self.styles.gm
            fx=self.fx+self.styles.gm

        if self.head and self.template.get('showHeader',True):
            head=self.replaceTokens(self.head,canv,doc)
            para=Paragraph(head,style=self.styles['header'])
            para.wrap(self.tw,self.styles.ph)
            para.drawOn(canv,hx,self.hy)
        if self.foot and self.template.get('showFooter',True):
            foot=self.replaceTokens(self.foot,canv,doc)
            para=Paragraph(foot,style=self.styles['footer'])
            para.wrap(self.tw,self.styles.ph)
            para.drawOn(canv,fx,self.fy)


def main():
    '''Parse command line and call createPdf with the correct data'''

    parser = OptionParser()
    parser.add_option('-o', '--output',dest='output',metavar='FILE'
                      ,help='Write the PDF to FILE')
                      
    def_ssheets=','.join([ os.path.expanduser(p) for p in \
    config.getValue("general","stylesheets","").split(',')])    
    parser.add_option('-s', '--stylesheets',dest='style',metavar='STYLESHEETS',default=def_ssheets,
                      help='A comma-separated list of custom stylesheets. Default="%s"'%def_ssheets)

    def_compressed=config.getValue("general","compressed",False)
    parser.add_option('-c', '--compressed',dest='compressed',action="store_true",default=def_compressed,
                      help='Create a compressed PDF. Default=%s'%def_compressed)
                      
    parser.add_option('--print-stylesheet',dest='printssheet',action="store_true",default=False,
                      help='Print the default stylesheet and exit')

    parser.add_option('--font-folder',dest='ffolder',metavar='FOLDER',
                      help='Search this folder for fonts. (Deprecated)')
                  
    def_fontpath=':'.join([ os.path.expanduser(p) for p in \
        config.getValue("general","font_path","").split(':')])
    parser.add_option('--font-path',dest='fpath',metavar='FOLDER:FOLDER:...:FOLDER',default=def_fontpath,
                      help='A colon-separated list of folders to search for fonts. Default="%s"'%def_fontpath)   

    
    def_lang=config.getValue("general","language","en_US")
    parser.add_option('-l','--language',metavar='LANG',default=def_lang,dest='language',
                      help='Language to be used for hyphenation. Default="%s"'%def_lang)
                      
    def_fit=config.getValue("general","fit_mode","shrink")
    parser.add_option('--fit-literal-mode',metavar='MODE',default=def_fit,dest='fitMode',
                      help='What todo when a literal is too wide. One of error,overflow,shrink,truncate. Default="%s"'%def_fit)
                      
    def_break=config.getValue("general","break_level",0)
    parser.add_option('-b','--break-level',dest='breaklevel',metavar='LEVEL',default=def_break,
                      help='Maximum section level that starts in a new page. Default: %d'%def_break)
                      
    parser.add_option('-q','--quiet',action="store_true",dest='quiet',default=False,
                      help='Print less information.')
    parser.add_option('-v','--verbose',action="store_true",dest='verbose',default=False,
                      help='Print debug information.')
    parser.add_option('--very-verbose',action="store_true",dest='vverbose',default=False,
                      help='Print even more debug information.')
    (options,args)=parser.parse_args()

    if options.quiet:
        log.setLevel(logging.CRITICAL)

    if options.verbose:
        log.setLevel(logging.INFO)

    if options.vverbose:
        log.setLevel(logging.DEBUG)

    if options.printssheet:
        print open(os.path.join(abspath(dirname(__file__)),'styles','styles.json')).read()
        sys.exit(0)

    filename = False
        
    if len(args) == 0:
        infile = sys.stdin
    elif len(args) > 1:
        log.critical('Usage: %s file.txt [ -o file.pdf ]', sys.argv[0])
        sys.exit(1)
    else:
        filename = args[0]
        infile=open(filename)
    
    if options.output:
        outfile=options.output
        if outfile =='-':
            outfile = sys.stdout
            options.compressed = False
            #we must stay quiet
            log.setLevel(logging.CRITICAL)
    else:
        if filename:
            if filename.endswith('.txt') or filename.endswith('.rst'):
                outfile = filename[:-4] + '.pdf'
            else:
                outfile=filename + '.pdf'
        else:
            outfile = sys.stdout
            options.compressed = False
            #we must stay quiet
            log.setLevel(logging.CRITICAL)
            #/reportlab/pdfbase/pdfdoc.py output can be a callable (stringio, stdout ...)            

    if options.style:
        ssheet=options.style.split(',')
    else:
        ssheet=[]
        
    fpath=[]
    if options.fpath:
        fpath=options.fpath.split(':')
    if options.ffolder:
        fpath.append(options.ffolder)

    RstToPdf(stylesheets = ssheet,
             language=options.language,
             breaklevel=int(options.breaklevel),
             fitMode=options.fitMode,
             fontPath=fpath).createPdf(text=infile.read(),
                                                 output=outfile,
                                                 compressed=options.compressed)

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
#$HeadURL$
#$LastChangedDate$
#$LastChangedRevision$
import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

long_description = (
    read('LICENSE.txt')
    + '\n' +
    'Detailed Documentation\n'
    '**********************\n'
    + '\n' +
    read('README.txt')
    + '\n' +
    'Contributors\n' 
    '************\n'

    
    + '\n' +
    read('Contributors.txt')
    + '\n' +
    'Change history\n'
    '**************\n'
    + '\n' + 
    read('CHANGES.txt')
    + '\n' +
   'Download\n'
    '********\n'
    )

install_requires = [
        'setuptools',
        'docutils',
        'PIL',
        'reportlab>=2.1',
        'Pygments',
        'simplejson'
        ]
        
tests_require = ['pyPdf',]
hyphenation_require = ['wordaxe',]

setup(
    name = "rst2pdf",
    version = "0.8",
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    package_data = {'rst2pdf': ['*/*.ttf','*.json']},
    namespace_packages=['rst2pdf'],
    include_package_data=True,
    # place to find an egg distrib of reportlab 2.1
    dependency_links = [
       #reportlab as egg
       "http://ftp.schooltool.org/schooltool/eggs/3.4",
       #wordaxe
       "http://sourceforge.net/project/platformdownload.php?group_id=105867",
       #PIL as egg
       "http://dist.repoze.org",
    ], 
    install_requires = install_requires,
    tests_require=tests_require,
    extras_require=dict(tests=tests_require,
                        hyphenation=hyphenation_require,),
    # metadata for upload to PyPI
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Documentation',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing',
        'Topic :: Utilities',
    ], 
    author = "Roberto Alsina",
    author_email = "ralsina at netmanagers dot com dot ar",
    description = "Convert restructured text to PDF via reportlab.",
    long_description=long_description,
    license = "MIT",
    keywords = "restructured convert rst pdf docutils pygments reportlab",
    url = "http://rst2pdf.googlecode.com",
    download_url = "http://code.google.com/p/rst2pdf/downloads/list",
    entry_points = {'console_scripts': ['rst2pdf = rst2pdf.createpdf:main']},
    test_suite = 'rst2pdf.tests.test_rst2pdf.test_suite',
)



'''Convert a pygments' generated CSS file into a rst2pdf stylesheet.

You can generate a pygments CSS file like this (replace murphy with whatever style you want)::
  
  pygmentize -S murphy -f html
  
'''

import sys,simplejson
from pygments.token import STANDARD_TYPES
dstyles={}
# First create a dumb stylesheet
for key in STANDARD_TYPES:
    dstyles["pygments-"+STANDARD_TYPES[key]]={'parent':'code'}


styles=[]
for line in open(sys.argv[1]):
    line=line.strip()
    sname="pygments-"+line.split(' ')[0][1:]
    style=dstyles.get(sname,{'parent':'code'})
    options=line.split('{')[1].split('}')[0].split(';')
    for option in options:
        option=option.strip()
        option,argument=option.split(':')
        if option=='color':
            style['textColor']=argument.strip()
        if option=='background-color':
            style['backColor']=argument.strip()
            
        # These two can come in any order
        if option=='font-weight' and argument=='bold':
            if 'fontName' in style and style['fontName']=='stdMonoItalic':
              style['fontName']='stdMonoBoldItalic'
            else:
              style['fontName']='stdMonoBold'
        if option=='font-style' and argument=='italic':
            if 'fontName' in style and style['fontName']=='stdBold':
              style['fontName']='stdMonoBoldItalic'
            else:
              style['fontName']='stdMonoItalic'
    styles.append([sname,style])
print simplejson.dumps({'styles':styles})



# -*- coding: utf-8 -*-
#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.8/rst2pdf/flowables.py $
#$LastChangedDate: 2008-09-11 23:03:46 -0300 (Thu, 11 Sep 2008) $
#$LastChangedRevision: 339 $
__docformat__ = 'reStructuredText'

from reportlab.platypus import *
from reportlab.platypus.doctemplate import *
from reportlab.lib.units import *
from reportlab.platypus.flowables import _listWrapOn, _FUZZ

from log import log
import styles

class MyIndenter(Indenter):
    '''I can't remember why this exists'''
    def draw(self):
        pass
    width=0
    height=0

class OutlineEntry(Flowable):
    '''Creates outline entries in the PDF TOC'''
    def __init__(self,label,text,level=0,snum=None):
        '''* label is a unique label.
           * text is the text to be displayed in the outline tree
           * level is the level, 0 is outermost, 1 is child of 0, etc.
        '''
        if label is None: # it happens
            self.label=text.replace(u'\xa0', ' ').strip(
                ).replace(' ', '_').encode('ascii', 'replace')
        else:
            self.label=label.strip()
        self.text=text.strip()
        self.level=int(level)
        self.snum=snum
        Flowable.__init__(self)

    def wrap(self,w,h):
        '''This takes no space'''
        return (0,0)

    def draw(self):
        self.canv.bookmarkPage(self.label)
        self.canv.sectName=self.text
        if self.snum is not None:
            self.canv.sectNum=self.snum
        else:
            self.canv.sectNum=""
        self.canv.addOutlineEntry(self.text,
                                  self.label,
                                  self.level, False)

    def __repr__(self):
        return "OutlineEntry (label=%s , text=%s , level=%d) \n"%(self.label,self.text,self.level)

class Separation(Flowable):
    """A simple <hr>-like flowable"""

    def wrap(self,w,h):
        self.w=w
        return (w,1*cm)

    def draw(self):
        self.canv.line(0,0.5*cm,self.w,0.5*cm)

class Reference(Flowable):
    '''A flowable to insert an anchor without taking space'''
    def __init__(self,refid):
        self.refid=refid
        Flowable.__init__(self)
        
    def wrap(self,w,h):
        '''This takes no space'''
        return (0,0)

    def draw(self):
        self.canv.bookmarkPage(self.refid)

    def repr(self):
        return "Anchor: %s"%self.refid

class DelayedTable(Flowable):
    '''A flowable that inserts a table for which it has the data.
    Needed so column widths can be determined after we know on what
    frame the table will be inserted, thus making the overal table width
    correct'''
    def __init__(self,data,colwidths,style):
        self.data=data
        self.colwidths=colwidths
        self.style=style
        self.t=None
    def wrap(self,w,h):
        # First create the table, with the widths from colwidths reinterpreted
        # if needed as percentages of frame/cell/whatever width w is.

        # Colwidths in ReST are relative: 10,10,10 means 33%,33%,33%
        # So, we need to calculate them relative to something
        # and we use the text width of the page. That sucks for
        # multicolumn pages
        # FIXME: make it work for multicolumn pages. No idea how, yet.
        _tw=w/sum(self.colwidths)
        colwidths=[ _w*_tw for _w in self.colwidths ]
        
        self.t=Table(self.data,colWidths=colwidths,style=TableStyle(self.style))
        return self.t.wrap(w,h)

    def split(self,w,h):
        return self.t.split(w,h)

    def drawOn(self,canvas,x,y,_sW=0):
        self.t.drawOn(canvas,x,y,_sW)

class MyPageBreak(FrameActionFlowable):
    def __init__(self, templateName=None):
        self.templateName=templateName

    def frameAction(self,frame):
        if not frame._atTop:
            frame._generated_content = [SetNextTemplate(self.templateName),PageBreak()]
        else:
            frame._generated_content = [SetNextTemplate(self.templateName)]

class SetNextTemplate(Flowable):
    """
    Sets canv.templateName when drawing.

    rst2pdf uses that to switch page templates.
    """

    def __init__(self, templateName=None):
        self.templateName=templateName
        Flowable.__init__(self)
        
    def draw(self):
        if self.templateName:
            self.canv.templateName=self.templateName

class SmartFrame(Frame):
    '''A (Hopefully) smarter frame object that knows how to
    handle a two-pass layout procedure (someday)'''

    def __init__(self, container,x1, y1, width,height, leftPadding=6, bottomPadding=6,
            rightPadding=6, topPadding=6, id=None, showBoundary=0,
            overlapAttachedSpace=None,_debug=None):
        self.container=container
        Frame.__init__(self,x1, y1, width,height, leftPadding, bottomPadding,
            rightPadding, topPadding, id, showBoundary,
            overlapAttachedSpace,_debug)

class FrameCutter(FrameActionFlowable):
    def __init__(self,dx,width,flowable,padding,lpad,floatLeft=True):
        self.width=width
        self.dx=dx
        self.f=flowable
        self.padding=padding
        self.lpad=lpad
        self.floatLeft=floatLeft
    def frameAction(self,frame):
        idx=frame.container.frames.index(frame)
        if self.floatLeft:
            if self.width-self.padding > 30: # Don´ t bother inserting a silly thin frame
                f1=SmartFrame(frame.container,
                              frame._x1+self.dx-2*self.padding,
                              frame._y2-self.f.height-3*self.padding,
                              self.width+2*self.padding,
                              self.f.height+3*self.padding,bottomPadding=0,topPadding=0,
                              leftPadding=self.lpad)
                f1._atTop=frame._atTop
                frame.container.frames.insert(idx+1,f1)

            if frame._height-self.f.height-2*self.padding >30: # Don't add silly thin frame
                frame.container.frames.insert(idx+2,SmartFrame(frame.container,
                                                               frame._x1,
                                                               frame._y1p,
                                                               self.width+self.dx,
                                                               frame._height-self.f.height-3*self.padding,topPadding=0))
        else:
            if self.width-self.padding > 30: # Don´ t bother inserting a silly thin frame
                f1=SmartFrame(frame.container,
                              frame._x1-self.width,
                              frame._y2-self.f.height-2*self.padding,
                              self.width,
                              self.f.height+2*self.padding,bottomPadding=0,topPadding=0,
                              rightPadding=self.lpad)
                f1._atTop=frame._atTop
                frame.container.frames.insert(idx+1,f1)
            if frame._height-self.f.height-2*self.padding >30:
                frame.container.frames.insert(idx+2,SmartFrame(frame.container,
                    frame._x1-self.width,
                    frame._y1p,
                    self.width+self.dx,
                    frame._height-self.f.height-2*self.padding,
                    topPadding=0))
                    
class Sidebar(FrameActionFlowable):
    def __init__(self,flowables,style):
        self.style=style
        self.width=self.style.width
        self.flowables=flowables

    def frameAction(self,frame):
        w=frame.container.styles.adjustUnits(self.width,frame.width)
        idx=frame.container.frames.index(frame)
        padding = self.style.borderPadding
        width=self.style.width
        self.style.padding=frame.container.styles.adjustUnits(str(padding),frame.width)
        self.style.width=frame.container.styles.adjustUnits(str(width),frame.width)
        self.kif=BoxedContainer(self.flowables,self.style)
        if self.style.float=='left':
            self.style.lpad = frame.leftPadding
            f1=SmartFrame(frame.container,
                          frame._x1,
                          frame._y1p,
                          w-2*self.style.padding,
                          frame._y-frame._y1p,
                          leftPadding=self.style.lpad,
                          rightPadding=0,
                          bottomPadding=0,
                          topPadding=0)
            f1._atTop=frame._atTop
            frame.container.frames.insert(idx+1,f1)
            frame._generated_content = [FrameBreak(),self.kif,
                FrameCutter(w,
                    frame.width-w,
                    self.kif,
                    padding,
                    self.style.lpad,
                    True),
                FrameBreak()]
        elif self.style.float=='right':
            self.style.lpad = frame.rightPadding
            frame.container.frames.insert(idx+1,SmartFrame(frame.container,
                                                       frame._x1+frame.width-self.style.width,
                                                       frame._y1p,
                                                       w,
                                                       frame._y-frame._y1p,
                                                       rightPadding=self.style.lpad,
                                                       leftPadding=0,
                                                       bottomPadding=0,
                                                       topPadding=0))
            frame._generated_content = [FrameBreak(),self.kif,
                FrameCutter(w,
                    frame.width-w,
                    self.kif,
                    padding,
                    self.style.lpad,
                    False),
                FrameBreak()]


class BoundByWidth(Flowable):
    '''Limits a list of flowables by width, but still lets them break
    over pages and frames'''
    
    def __init__(self, maxWidth, content=[], style=None, mode=None):
        self.maxWidth=maxWidth
        self.content=content
        self.style=style
        self.mode=mode
        if self.style:
            self.pad = self.style.borderPadding+self.style.borderWidth+.1
        else:
            self.pad=0
        Flowable.__init__(self)

    def identity(self, maxLen=None):
        return "<%s at %s%s%s> size=%sx%s" % (self.__class__.__name__, hex(id(self)), self._frameName(),
                getattr(self,'name','') and (' name="%s"'% getattr(self,'name','')) or '',
                getattr(self,'maxWidth','') and (' maxWidth=%s'%str(getattr(self,'maxWidth',0))) or '',
                getattr(self,'maxHeight','')and (' maxHeight=%s' % str(getattr(self,'maxHeight')))or '')

    def wrap(self,availWidth,availHeight):
        '''If we need more width than we have, complain, keep a scale'''
        if self.style:
            self.pad = self.style.borderPadding+self.style.borderWidth+.1
        else:
            self.pad=0
        maxWidth = float(min(styles.adjustUnits(self.maxWidth,availWidth) or availWidth,availWidth))
        self.maxWidth=maxWidth
        maxWidth -= 2*self.pad
        self.width, self.height = _listWrapOn(self.content,maxWidth,self.canv)
        self.scale=1.0
        if self.width > maxWidth:
            log.warning("BoundByWidth too wide to fit in frame: %s",self.identity())
            if self.mode=='shrink':
                self.scale=(maxWidth+2*self.pad)/(self.width+2*self.pad)
                self.height=self.height*self.scale
            #self.width=maxWidth
        # No need to warn here, it will just split
        #if self.height+2*self.pad*self.scale > availHeight:
            #log.warning("BoundByWidth too tall to fit in frame: %s",self.identity())
        return self.width, self.height+2*self.pad*self.scale

    def split(self,availWidth,availHeight):
        if len(self.content)>1:
            # Try splitting in our individual elements
            return [ BoundByWidth(self.maxWidth,[f],self.style) for f in self.content ]
        else: # We need to split the only element we have
            return [ BoundByWidth(self.maxWidth,[f],self.style) for f in self.content[0].split(availWidth-2*self.pad,availHeight-2*self.pad) ]

    def draw(self):
        '''we simulate being added to a frame'''
        canv=self.canv
        canv.saveState()
        x=canv._x
        y=canv._y
        _sW=0
        scale=self.scale
        content=None
        aW=None
        #, canv, x, y, _sW=0, scale=1.0, content=None, aW=None):
        pS = 0
        if aW is None: aW = self.width
        aW = scale*(aW+_sW)
        if content is None:
            content = self.content
        y += (self.height+self.pad)/scale
        x += self.pad
        for c in content:
            w, h = c.wrapOn(canv,aW,0xfffffff)
            if (w<_FUZZ or h<_FUZZ) and not getattr(c,'_ZEROSIZE',None): continue
            if c is not content[0]: h += max(c.getSpaceBefore()-pS,0)
            y -= h
            canv.saveState()
            if self.mode=='shrink':
                canv.scale(scale,scale)
            elif self.mode=='truncate':
                p=canv.beginPath()
                p.rect(x-self.pad,y-self.pad,self.maxWidth,self.height+2*self.pad)
                canv.clipPath(p,stroke=0)
            c.drawOn(canv,x,y,_sW=aW-w)
            canv.restoreState()
            if c is not content[-1]:
                pS = c.getSpaceAfter()
                y -= pS
        canv.restoreState()

class BoxedContainer(BoundByWidth):
    def __init__(self, content, style, mode='shrink'):
        BoundByWidth.__init__(self,style.width, content, mode=mode, style=None)
        self.style=style

    def draw(self):
        canv=self.canv
        canv.saveState()
        x=canv._x
        y=canv._y
        _sW=0
        lw=0
        if self.style and self.style.borderWidth >0:
            lw=self.style.borderWidth
            canv.setLineWidth(self.style.borderWidth)
            canv.setStrokeColor(self.style.borderColor)
        if self.style and self.style.backColor:
            canv.setFillColor(self.style.backColor)
        if self.style and self.style.padding:
            self.padding=self.style.padding
        else:
            self.padding=0
        self.padding+=lw
        p = canv.beginPath()
        p.rect(x, y, self.width+2*self.padding,self.height+2*self.padding)
        canv.drawPath(p,stroke=1,fill=1)
        canv.restoreState()
        BoundByWidth.draw(self)

if reportlab.Version == '2.1':

    import reportlab.platypus.paragraph as pla_para

    ################Ugly stuff below

    def _do_post_text(i, t_off, tx):
        '''From reportlab's paragraph.py, patched to avoid underlined links'''
        xs = tx.XtraState
        leading = xs.style.leading
        ff = 0.125*xs.f.fontSize
        y0 = xs.cur_y - i*leading
        y = y0 - ff
        ulc = None
        for x1,x2,c in xs.underlines:
            if c!=ulc:
                tx._canvas.setStrokeColor(c)
                ulc = c
            tx._canvas.line(t_off+x1, y, t_off+x2, y)
        xs.underlines = []
        xs.underline=0
        xs.underlineColor=None

        ys = y0 + 2*ff
        ulc = None
        for x1,x2,c in xs.strikes:
            if c!=ulc:
                tx._canvas.setStrokeColor(c)
                ulc = c
            tx._canvas.line(t_off+x1, ys, t_off+x2, ys)
        xs.strikes = []
        xs.strike=0
        xs.strikeColor=None

        yl = y + leading
        for x1,x2,link in xs.links:
            # This is the bad line
            #tx._canvas.line(t_off+x1, y, t_off+x2, y)
            _doLink(tx, link, (t_off+x1, y, t_off+x2, yl))
        xs.links = []
        xs.link=None

    # Look behind you! A three-headed monkey!
    pla_para._do_post_text.func_code = _do_post_text.func_code

    ############### End of the ugly
